/*-----------------------------------------------------------------------*/
/* files.c --- file handling routines for xcircuit			 */
/* Copyright (c) 2001  Tim Edwards, Johns Hopkins University        	 */
/*-----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(__DARWIN__)
#include <malloc.h>
#endif
#include <time.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xw/TextEdit.h>   /* for XwTextCopyBuffer() */

/*------------------------------------------------------------------------*/
/* Local includes                                                         */
/*------------------------------------------------------------------------*/

#include "colordefs.h"
#include "xcircuit.h"

/*----------------------------------------------------------------------*/
/* Function prototype declarations                                      */
/*----------------------------------------------------------------------*/
#include "prototypes.h"

/*------------------------------------------------------------------------*/
/* Useful (local) defines						  */
/*------------------------------------------------------------------------*/

#define OUTPUTWIDTH 80	/* maximum text width of output */

#define S_OBLIQUE 13	/* position of Symbol-Oblique in font array */

/* type checks */

#define IS_POLYGON(a) 	((*a)->type == POLYGON)
#define IS_LABEL(a)	((*a)->type == LABEL)
#define IS_OBJINST(a)	((*a)->type == OBJECT)
#define IS_ARC(a)	((*a)->type == ARC)
#define IS_SPLINE(a)	((*a)->type == SPLINE)
#define IS_PATH(a)	((*a)->type == PATH)


/*------------------------------------------------------------------------*/
/* External Variable definitions                                          */
/*------------------------------------------------------------------------*/

extern char _STR2[250], _STR[150];
extern Globaldata xobjs;
extern Clientdata areastruct;
extern fontinfo *fonts;
extern short fontcount;
extern Display *dpy;
extern Window win;
extern short eventmode;
extern short beeper;
extern int *appcolors;
extern int number_colors;
extern colorindex *colorlist;
extern objectpair *pushlist;

/*------------------------------------------------------*/
/* Global variable definitions				*/
/*------------------------------------------------------*/

float version;

/*------------------------------------------------------*/
/* Free memory allocated to a label string		*/
/*------------------------------------------------------*/

void freelabel(stringpart *string)
{
   stringpart *strptr = string, *tmpptr;

   while (strptr != NULL) {
      if (strptr->type == TEXT_STRING)
         free(strptr->data.string);
      tmpptr = strptr->nextpart;
      free(strptr);
      strptr = tmpptr;
   }
}

/*------------------------------------------------------*/
/* Free memory for a single object			*/
/*------------------------------------------------------*/

void free_single(genericptr *genobj)
{
   int i;
   objinstptr geninst;

   if (IS_POLYGON(genobj)) free(((polyptr)(*genobj))->points);
   else if (IS_LABEL(genobj)) freelabel(((labelptr)(*genobj))->string);
   else if (IS_PATH(genobj)) free(((pathptr)(*genobj))->plist);
   else if (IS_OBJINST(genobj)) {
      geninst = TOOBJINST(genobj);
      if (geninst->params != NULL) {
         for (i = 0; i < geninst->thisobject->num_params; i++) {
	    if (geninst->params[i] != NULL) {
	       if (geninst->params[i]->type == XC_STRING)
		  freelabel(geninst->params[i]->parameter.string);
	       free (geninst->params[i]);
	    }
	 }
         free(geninst->params);
      }
   }
   free(*genobj);
}

/*---------------------------------------------------------*/
/* Reset an object structure by freeing all alloc'd memory */
/*---------------------------------------------------------*/

void reset(objectptr localdata, short mode)
{
   if (localdata->parts > 0) {
      genericptr *genobj;
      if (mode != SAVE) {
	 for (genobj = localdata->plist; genobj < localdata->plist
	        + localdata->parts; genobj++)
	    free_single(genobj);
      }
      free(localdata->plist);
      
      removeparams(localdata);

      initmem(localdata);
      if (mode == DESTROY) free(localdata->plist);
   }
}

/*---------------------------------------------------------*/

void topreset()
{
   short rpage;
   objectptr resetpage;

   /* free alloc'd filename but */
   /* don't free it if another page points to it */

   if (xobjs.pagelist[areastruct.page]->filename != NULL) {
      for (rpage = 0; rpage < xobjs.pages; rpage++) {
         resetpage = xobjs.pagelist[rpage]->pageobj;
         if (resetpage != NULL) {
            if ((rpage != areastruct.page) && xobjs.pagelist[rpage]->filename ==
	        xobjs.pagelist[areastruct.page]->filename) break;
         }
      }
      if (rpage == xobjs.pages) free(xobjs.pagelist[areastruct.page]->filename);
   }
   if (xobjs.pagelist[areastruct.page]->background.name != NULL)
      free(xobjs.pagelist[areastruct.page]->background.name);

   rpage = areastruct.page;

   if (areastruct.selects > 0) {
      free(areastruct.selectlist);
      areastruct.selects = 0;
   }

   /* New pages pick up their properties from page 0, which can be changed */
   /* from the .xcircuitrc file on startup (or loaded from a script).	   */
   /* Thanks to Norman Werner (norman.werner@student.uni-magdeburg.de) for */
   /* pointing out this more obvious way of doing the reset, and providing */
   /* a patch.								   */

   xobjs.pagelist[rpage]->wirewidth = xobjs.pagelist[0]->wirewidth;
   xobjs.pagelist[rpage]->orient = xobjs.pagelist[0]->orient;
   xobjs.pagelist[rpage]->pmode = xobjs.pagelist[0]->pmode;
   xobjs.pagelist[rpage]->outscale = xobjs.pagelist[0]->outscale;
   xobjs.pagelist[rpage]->filename = (char *)malloc
		((1 + strlen(xobjs.pagelist[rpage]->pageobj->name)) * sizeof(char));
   xobjs.pagelist[rpage]->background.name = (char *)NULL;

   strcpy(xobjs.pagelist[rpage]->filename, xobjs.pagelist[rpage]->pageobj->name);
   xobjs.pagelist[rpage]->drawingscale.x = xobjs.pagelist[0]->drawingscale.x;
   xobjs.pagelist[rpage]->drawingscale.y = xobjs.pagelist[0]->drawingscale.y;
   xobjs.pagelist[rpage]->gridspace = xobjs.pagelist[0]->gridspace;
   xobjs.pagelist[rpage]->snapspace = xobjs.pagelist[0]->snapspace;

   if (xobjs.pagelist[0]->coordstyle == CM) {
      xobjs.pagelist[rpage]->coordstyle = CM;
      xobjs.pagelist[rpage]->pagesize.x = 595;
      xobjs.pagelist[rpage]->pagesize.y = 842;  /* A4 */
   }
   else {
      xobjs.pagelist[rpage]->coordstyle = FRAC_INCH;
      xobjs.pagelist[rpage]->pagesize.x = 612;	/* letter */
      xobjs.pagelist[rpage]->pagesize.y = 792;
   }
}

/*---------------------------------------------------------*/

void initmem(objectptr localdata)
{
   localdata->parts = 0;
   localdata->plist = (genericptr *)malloc(sizeof(genericptr));
   localdata->hidden = False;
   localdata->params = NULL;
   localdata->num_params = 0;

   localdata->viewscale = 0.5;
   localdata->pcorner.x = -areastruct.width;
   localdata->pcorner.y = -areastruct.height; 
   localdata->width = 0;
   localdata->height = 0;
   localdata->lowerleft.x = 0;
   localdata->lowerleft.y = 0;

#ifdef SCHEMA
   localdata->schemtype = SCHEMATIC;
   localdata->symschem = NULL;
   localdata->netlist = NULL;
   localdata->portlist = NULL;
   localdata->calllist = NULL;
   localdata->traversed = 0;
   localdata->devname = NULL;

   localdata->width2 = 0;
   localdata->height2 = 0;
   localdata->lleft2.x = 0;
   localdata->lleft2.y = 0;
#endif
}

/*--------------------------------------------------------------*/
/* Exhaustively compare the contents of two objects and return  */
/* true if equivalent, false if not.				*/
/*--------------------------------------------------------------*/

Boolean elemcompare(genericptr *compgen, genericptr *gchk)
{
   Boolean bres;
   switch((*compgen)->type) {
      case(ARC):
	 bres = (TOARC(compgen)->position.x == TOARC(gchk)->position.x &&
            TOARC(compgen)->position.y == TOARC(gchk)->position.y &&
	    TOARC(compgen)->style == TOARC(gchk)->style &&
	    TOARC(compgen)->width == TOARC(gchk)->width &&
	    abs(TOARC(compgen)->radius) == abs(TOARC(gchk)->radius) &&
	    TOARC(compgen)->yaxis  == TOARC(gchk)->yaxis &&
	    TOARC(compgen)->angle1 == TOARC(gchk)->angle1 &&
	    TOARC(compgen)->angle2 == TOARC(gchk)->angle2);
	 break;
      case(SPLINE):
	 bres = (TOSPLINE(compgen)->style == TOSPLINE(gchk)->style &&
	    TOSPLINE(compgen)->width == TOSPLINE(gchk)->width &&
	    TOSPLINE(compgen)->ctrl[0].x == TOSPLINE(gchk)->ctrl[0].x &&
	    TOSPLINE(compgen)->ctrl[0].y == TOSPLINE(gchk)->ctrl[0].y &&
	    TOSPLINE(compgen)->ctrl[1].x == TOSPLINE(gchk)->ctrl[1].x &&
	    TOSPLINE(compgen)->ctrl[1].y == TOSPLINE(gchk)->ctrl[1].y &&
	    TOSPLINE(compgen)->ctrl[2].x == TOSPLINE(gchk)->ctrl[2].x &&
	    TOSPLINE(compgen)->ctrl[2].y == TOSPLINE(gchk)->ctrl[2].y &&
	    TOSPLINE(compgen)->ctrl[3].x == TOSPLINE(gchk)->ctrl[3].x &&
	    TOSPLINE(compgen)->ctrl[3].y == TOSPLINE(gchk)->ctrl[3].y);
	 break;
      case(POLYGON): {
	 int i;
	 if (TOPOLY(compgen)->style == TOPOLY(gchk)->style &&
	       TOPOLY(compgen)->width == TOPOLY(gchk)->width &&
	       TOPOLY(compgen)->number == TOPOLY(gchk)->number) {
	    for (i = 0; i < TOPOLY(compgen)->number; i++) {
	       if (TOPOLY(compgen)->points[i].x != TOPOLY(gchk)->points[i].x
		     || TOPOLY(compgen)->points[i].y != TOPOLY(gchk)->points[i].y)
		  break;
	    }
	    bres = (i == TOPOLY(compgen)->number);
	 }
	 else bres = False;
	 }break;
   }
   return bres;
}

/*--------------------------------------------------------------*/
/* Compare any element with any other element.			*/
/*--------------------------------------------------------------*/

Boolean compare_single(genericptr *compgen, genericptr *gchk)
{
   Boolean bres = False;

   if ((*gchk)->type == (*compgen)->type) {
      switch((*compgen)->type) {
	 case(OBJECT):{
	    objinst *newobj = TOOBJINST(compgen);
	    objinst *oldobj = TOOBJINST(gchk);
	    bres = (newobj->position.x == oldobj->position.x &&
	    		newobj->position.y == oldobj->position.y &&
	     		newobj->rotation == oldobj->rotation &&
	     		newobj->scale == oldobj->scale &&
	     		newobj->thisobject == oldobj->thisobject &&
			newobj->thisobject->num_params
				== oldobj->thisobject->num_params);	/* more! */
	    } break;
	 case(LABEL):
	    bres = (TOLABEL(compgen)->position.x == TOLABEL(gchk)->position.x &&
	    		TOLABEL(compgen)->position.y == TOLABEL(gchk)->position.y &&
	     		TOLABEL(compgen)->rotation == TOLABEL(gchk)->rotation &&
	     		TOLABEL(compgen)->scale == TOLABEL(gchk)->scale &&
	     		TOLABEL(compgen)->justify == TOLABEL(gchk)->justify &&
#ifdef SCHEMA
			((TOLABEL(compgen)->pin == TOLABEL(gchk)->pin) ||
				!areastruct.schemon) &&
#endif
	     		!stringcomp(TOLABEL(compgen)->string, TOLABEL(gchk)->string));
	    break;
	 case(PATH): /* elements *must* be in same order for a path */
	    bres = (TOPATH(compgen)->parts == TOPATH(gchk)->parts &&
			TOPATH(compgen)->style == TOPATH(gchk)->style &&
			TOPATH(compgen)->width == TOPATH(gchk)->width);
	    if (bres) {
	       genericptr *pathchk, *gpath;
	       for (pathchk = TOPATH(compgen)->plist, gpath =
			   TOPATH(gchk)->plist; pathchk < TOPATH(compgen)->plist
			   + TOPATH(compgen)->parts; pathchk++, gpath++) {
		  if (!elemcompare(pathchk, gpath)) bres = False;
	       }
	    }
	    break;
	 case(ARC): case(SPLINE): case(POLYGON):
	    bres = elemcompare(compgen, gchk);
	    break;
      }
   }
   return bres;
}

/*--------------------------------------------------------------------*/

short objcompare(objectptr obja, objectptr objb)
{
   genericptr *compgen, *glist, *gchk, *remg;
   short      csize;
   Boolean    bres;

   /* quick check on equivalence of number of objects */

   if (obja->parts != objb->parts) return False;

   /* check equivalence of parameters */

   if (obja->params == NULL && objb->params != NULL) return False;
   else if (obja->params != NULL && objb->params == NULL) return False;
   else if (obja->params != NULL || objb->params != NULL) {
      if (obja->num_params != objb->num_params) return False;
      else {
	 short i;
	 for (i = 0; i < obja->num_params; i++) {
	    oparamptr opsa = *(obja->params + i);
	    oparamptr opsb = *(objb->params + i);
	    if (opsa->type != opsb->type) return False;
	    if (opsa->type == XC_STRING && 
		 	stringcomp(opsa->parameter.string, opsb->parameter.string))
	       return False;
	    else if (opsa->type == XC_SHORT &&
		  opsa->parameter.svalue != opsb->parameter.svalue)
	       return False;
	 }
      }
   }

   /* For the exhaustive check we must match component for component. */
   /* Best not to assume that elements are in same order for both.    */

   csize = obja->parts;

   glist = (genericptr *)malloc(csize * sizeof(genericptr));
   for (compgen = objb->plist; compgen < objb->plist + csize; compgen++)
      (*(glist + (int)(compgen - objb->plist))) = *compgen;
   for (compgen = obja->plist; compgen < obja->plist + obja->parts;
	 compgen++) {
      bres = False;
      for (gchk = glist; gchk < glist + csize; gchk++) {
	 if ((*compgen)->color == (*gchk)->color)
	    bres = compare_single(compgen, gchk);
	 if (bres) {
	   csize--;
	   for (remg = gchk; remg < glist + csize; remg++)
               *remg = *(remg + 1);
           break;
	 }
      }
   }
   free(glist);
   if (csize != 0) return False;

   return(True);
}

/*------------------------*/
/* scale renormalization  */
/*------------------------*/

float getpsscale(float value, short page)
{
   if (xobjs.pagelist[page]->coordstyle == FRAC_INCH ||
 	 xobjs.pagelist[page]->coordstyle == DEC_INCH)
      return (value * INCHSCALE);
   else
      return (value * CMSCALE);
}

/*---------------------------------------------------------------*/
/* Keep track of columns of output and split lines when too long */
/*---------------------------------------------------------------*/

void dostcount(FILE *ps, short *count, short addlength)
{
   *count += addlength;
   if (*count > OUTPUTWIDTH) {
      *count = addlength;
      fprintf(ps, "\n");
   }
}

/*-----------------------------------------------*/
/* Check parameters for pointer to a given point */
/*-----------------------------------------------*/

#define xymat(z) \
    varcheck(ps, &(z.x), localdata, xaddin, &stcount); \
    varcheck(ps, &(z.y), localdata, yaddin, &stcount)
#define xypmat(z) \
    varcheck(ps, &(z->x), localdata, xaddin, &stcount); \
    varcheck(ps, &(z->y), localdata, yaddin, &stcount)
#define xyfmat(z) \
   fprintf(ps, "%hd %hd ", (int)z.x + xaddin, (int)z.y + yaddin);
  
/*----------------------------------------------------------------------*/
/* Write a numerical value as a string to _STR, making a parameter	*/
/* substitution if appropriate (e.g., "v1", "v2", etc.).		*/
/*									*/
/* Note: As of version 2.2.2, there is still no procedure for integer	*/
/* integer variable substitutions.					*/
/*----------------------------------------------------------------------*/

void varcheck(FILE *ps, short *value, objectptr localdata, int addin, short *stptr)
{
   sprintf(_STR, "%d ", (int)(*value) + addin);
   dostcount (ps, stptr, strlen(_STR));
   fputs(_STR, ps);
}

/*------------------------*/
/* Load a PostScript file */
/*------------------------*/

#define LOAD_MODES 4

void getfile(Widget button, void *mode, caddr_t nulldata)
{
   buttonsave *savebutton;
   static void ((*runprog[LOAD_MODES])()) =
	{startloadfile, importfile, loadbackground, execscript};
   static char *substr[LOAD_MODES] = 
	{"load", "import", "render", "execute"};
   char promptstr[25];

   if (is_page(objectdata) == -1) {
      Wprintf("Can only read file into top-level page!");
      return;
   }
   else if ((int)mode >= LOAD_MODES) {
      Wprintf("Unknown mode passed to routine getfile()\n");
      return;
   }
   sprintf(promptstr, "Select file to %s:", substr[(int)mode]);
   savebutton = (buttonsave *)malloc(sizeof(buttonsave));
   getgeneric(savebutton, button, getfile, mode);
   popupprompt(button, promptstr, "\0", runprog[(int)mode],
		savebutton, True);
}

/*--------------------------------------------------------------*/
/* Tilde ('~') expansion in file name (stored in _STR)		*/
/*--------------------------------------------------------------*/

Boolean tilde_expand(char *filename)
{
   struct passwd *passwd;
   char *username = NULL, *expanded, *sptr;

   if (*filename == '~') {
      sptr = filename + 1;
      if (*sptr == '/' || *sptr == ' ' || *sptr == '\0')
	 username = getenv("HOME");
      else {
	 for (; *sptr != '/' && *sptr != '\0'; sptr++);
	 if (*sptr == '\0') *(sptr + 1) = '\0';
	 *sptr = '\0';

	 passwd = getpwnam(filename + 1);
	 if (passwd != NULL)
	    username = passwd->pw_dir;

	 *sptr = '/';
      }
      if (username != NULL) {
	 expanded = (char *)malloc(strlen(username) +
		strlen(filename));
	 strcpy(expanded, username);
         strcat(expanded, sptr);
	 strcpy(filename, expanded);
         free(expanded);
      }
      return True;
   }
   return False;
} 

/*---------------------------------------------------------*/

Boolean nextfilename()	/* extract next filename from _STR2 into _STR */
{
   char *cptr, *slptr;

   sprintf(_STR, "%.149s", _STR2);
   if ((cptr = strrchr(_STR2, ',')) != NULL) {
      slptr = strrchr(_STR, '/');
      if (slptr == NULL || ((slptr - _STR) > (cptr - _STR2))) slptr = _STR - 1;
      sprintf(slptr + 1, "%s", cptr + 1); 
      *cptr = '\0';
      return True;
   }
   else return False;
}

/*---------------------------------------------------------*/

void loadfontlib()
{
   loadlibrary(FONTLIB);
}

/*------------------------------------------------------*/
/* Handle library loading and refresh current page if 	*/
/* it is a library page that just changed.		*/
/*------------------------------------------------------*/

void loadglib(Boolean lflag, short ilib, short tlib)
{
   while (nextfilename()) {
      if (lflag)
	 lflag = False;
      else
         ilib = createlibrary();
      loadlibrary(ilib);
      if (ilib == tlib) zoomview(NULL, NULL, NULL);
   }
   if (lflag)
      lflag = False;
   else
      ilib = createlibrary();
   loadlibrary(ilib);
   if (ilib == tlib) zoomview(NULL, NULL, NULL);
}

/*------------------------------------------------------*/
/* Load new library:  Create new library page and load	*/
/* to it.						*/
/*------------------------------------------------------*/

void loadulib()
{
   loadglib(False, (short)0, (short)is_library(objectdata) + LIBRARY);
}

/*-----------------------------------------------------------*/
/* Add to library:  If on library page, add to that library. */
/* Otherwise, create a new library page and load to it.	     */
/*-----------------------------------------------------------*/

void loadblib()
{
   short ilib, tlib;
   Boolean lflag = True;

   /* Flag whether current page is a library page or not */

   if ((tlib = is_library(objectdata)) < 0) {
      ilib = LIBRARY;
      lflag = False;
   }
   else
      ilib = tlib + LIBRARY;

   loadglib(lflag, ilib, tlib + LIBRARY);
}

/*---------------------------------------------------------*/

void getlib(Widget button, caddr_t clientdata, caddr_t nulldata)
{
   buttonsave *savebutton = (buttonsave *)malloc(sizeof(buttonsave));
   getgeneric(savebutton, button, getlib, NULL);
   popupprompt(button, "Enter library to load:", "\0", loadblib, savebutton,
	True);
}

/*---------------------------------------------------------*/

void getuserlib(Widget button, caddr_t clientdata, caddr_t nulldata)
{
   buttonsave *savebutton = (buttonsave *)malloc(sizeof(buttonsave));
   getgeneric(savebutton, button, getuserlib, NULL);
   popupprompt(button, "Enter library to load:", "\0", loadulib, savebutton,
	True);
}

/*------------------------------------------------------*/
/* Load a library page (given in parameter "mode") and	*/
/* rename the library page to match the library name as */
/* found in the file header.				*/
/*------------------------------------------------------*/

void loadlibrary(short mode)
{
   FILE *ps;
   char inname[150], temp[150], keyword[30], percentc;

   sscanf(_STR, "%s", inname);

   ps = fopen(inname, "r");
   if (ps == NULL) {
      char *inptr;

      /* try adding a .lps */
      
      if ((inptr = strrchr(inname, '/')) == NULL) inptr = inname;
      if (strchr(inptr, '.') == NULL) {
	 sprintf(inname, "%s.lps", _STR);
         ps = fopen(inname, "r");
      }
      if (ps == NULL) {

         /* if not found in cwd, look for environment variable  */
	 /* "XCIRCUIT_LIB_DIR" defined			 	*/
	 /* (Thanks to Ali Moini, U. Adelaide, S. Australia)	*/

	 char *tmp_s = getenv((const char *)"XCIRCUIT_LIB_DIR");

	 if (tmp_s != NULL) {
	    sprintf(inname, "%s/%s", tmp_s, _STR);
      	    ps = fopen(inname, "r");
      	    if (ps == NULL) {
	       sprintf(inname, "%s/%s.lps", tmp_s, _STR);
	       ps = fopen(inname, "r");
	    }
            if (ps == NULL && mode == FONTLIB) {
	       /* printf("looking in %s/fonts\n", tmp_s); */
	       sprintf(inname, "%s/fonts/%s", tmp_s, _STR);
	       ps = fopen(inname, "r");
	       if (ps == NULL) {
	          sprintf(inname, "%s/fonts/%s.lps", tmp_s, _STR);
	          ps = fopen(inname, "r");
	       }
	       if (ps == NULL) printf("not found there.\n");
            }
	 }

	 /* last resort:  hard-coded directory BUILTINS_DIR */

	 if (ps == NULL) {
	    sprintf(inname, "%s/%s", BUILTINS_DIR, _STR);
      	    ps = fopen(inname, "r");
      	    if (ps == NULL) {
	       sprintf(inname, "%s/%s.lps", BUILTINS_DIR, _STR);
	       ps = fopen(inname, "r");
	    }
            if (ps == NULL && mode == FONTLIB) {
	       /* printf("looking in %s/fonts\n", BUILTINS_DIR); */
	       sprintf(inname, "%s/fonts/%s", BUILTINS_DIR, _STR);
	       ps = fopen(inname, "r");
	       if (ps == NULL) {
	          sprintf(inname, "%s/fonts/%s.lps", BUILTINS_DIR, _STR);
	          ps = fopen(inname, "r");
	       }
	       if (ps == NULL) printf("not found there.\n");
	    }
	    if (ps == NULL) {
               Wprintf("No library file found.");
               return;
	    }
	 }
      }
   }

   /* current version is PROG_VERSION;  however, all libraries newer than */
   /* version 2.0 require "Version" in the header.  So unnumbered	  */
   /* libraries may be assumed to be version 1.9 or earlier.		  */

   version = 1.9;
   for(;;) {
      if (fgets(temp, 149, ps) == NULL) {
         Wprintf("Error in library.");
         return;
      }
      sscanf(temp, "%c %29s", &percentc, keyword);

      /* Commands in header are PostScript comments (%) */
      if (percentc == '%') {

         /* Rename the library page if the library name is found in the header 	*/
	 /* and this library is not a font description.				*/
         if ((mode != FONTLIB) && !strcmp(keyword, "Library")) {
	    char *cptr, *nptr;
	    cptr = strchr(temp, ':');
	    if (cptr != NULL) {
	       /* Don't write terminating newline to the object's name string */
	       if ((nptr = strchr(cptr + 2, '\n')) != NULL) *nptr = '\0';
	       if (xobjs.userlibs[mode - LIBRARY].number == 0) {
	          sprintf(xobjs.libtop[mode]->name, "Library: %.79s", cptr + 2);
		  renamelib(mode);
	       }
	    }
         }

         /* This comment gives the Xcircuit version number */
	 else if (!strcmp(keyword, "Version:")) {
	    float tmpv;
	    if (sscanf(temp, "%*c %*s %f", &tmpv) > 0) version = tmpv;
	 }

         /* This PostScript comment marks the end of the file header */
         else if (!strcmp(keyword, "XCircuitLib")) break;
      }
   }
   objectread(ps, objectdata, 0, 0, mode, temp, DEFAULTCOLOR);

   if (mode != FONTLIB) composelib(mode);

   /* printf("Loaded library %s\n", inname); */
   sprintf(_STR, "Loaded library %s", inname);
   Wprintf(_STR);

   version = PROG_VERSION;
}

/*---------------------------------------------------------*/

void startloadfile()
{
   short firstpage = areastruct.page;

   while (nextfilename()) {
      loadfile(0);

      /* find next undefined page */

      while(areastruct.page < xobjs.pages &&
	   xobjs.pagelist[areastruct.page]->pageobj != NULL) areastruct.page++;
      changepage(areastruct.page);
   }
   loadfile(0);

   /* Display the first page loaded */

   newpage(firstpage);
   zoomview(NULL, NULL, NULL);

#ifdef SCHEMA
   setsymschem();
#endif
}

/*------------------------------------------------------*/
/* Import an xcircuit file onto the current page	*/
/*------------------------------------------------------*/

void importfile()
{
   while (nextfilename()) loadfile(1);
   loadfile(1);
   zoomview(NULL, NULL, NULL);
}

/*--------------------------------------------------------------*/
/* Load an xcircuit file into xcircuit				*/
/*--------------------------------------------------------------*/

void loadfile(short mode)
{
   FILE *ps;
   char inname[250], temp[150], keyword[30], percentc, *pdchar;
   char teststr[50], teststr2[20], pagestr[100];
   short offx, offy, multipage, page, temppmode = 0, oldpage;
   float tmpfl;
   XPoint pagesize;

   /* First, if we're in catalog mode, return with error */

   if (eventmode == CATALOG_MODE) {
      Wprintf("Cannot load file from library window");
      return;
   }

   /* Do tilde expansion on filename */

   tilde_expand(_STR);

#ifdef LGF
   /* quick check for .lgf or .lfo file */
   if ((pdchar = strchr(_STR, '.')) != NULL)
      if (!strcmp(pdchar + 1, "lgf") || !strcmp(pdchar + 1, "lfo")) {
	 loadlgf(mode);
         return;
      }
#endif

   sscanf(_STR, "%s", inname);
   ps = fopen(inname, "r");
   if (ps == NULL) {
      sprintf(inname, "%s.ps", _STR);
      ps = fopen(inname, "r");
      if (ps == NULL) {

	 /* Could possibly be a library file? */

         if ((pdchar = strchr(_STR, '.')) == NULL) {
	    sprintf(inname, "%s.lps", _STR);
	    if ((ps = fopen(inname, "r")) != NULL) {
	       fclose(ps);
	       loadlibrary(USERLIB);
	       return;
	    }

#ifdef LGF
	    sprintf(inname, "%s.lgf", _STR);
	    if ((ps = fopen(inname, "r")) != NULL) {
	       fclose(ps);
	       loadlgf(mode);
	       return;
	    }

	    sprintf(inname, "%s.lfo", _STR);
	    if ((ps = fopen(inname, "r")) != NULL) {
	       fclose(ps);
	       loadlgf(mode);
	       return;
	    }
#endif
	 }

	 /* What to do if no file was found:				  */
	 /* If extension was .lgf or .lfo, just flag an error.  	  */
         /* Otherwise, if we're on an empty page, go ahead and rename it. */

#ifdef LGF

         else {
	    if (!strcmp(pdchar + 1, "lgf") || !strcmp(pdchar + 1, "lfo")) {
	       sprintf(temp, "Error: Can't find lgf-format file %s", _STR);
               Wprintf(temp);
	       return;
	    }
	 }
#endif

         if (objectdata->parts == 0 && (mode == 0)) {

	    /* Check for file extension, and remove if "ps". */

            if ((pdchar = strchr(_STR, '.')) != NULL)
		if (!strcmp(pdchar + 1, "ps")) *pdchar = '\0';

            xobjs.pagelist[areastruct.page]->filename = (char *)realloc(
		 xobjs.pagelist[areastruct.page]->filename, (strlen(_STR) + 1)
		 * sizeof(char));
            sprintf(xobjs.pagelist[areastruct.page]->filename, "%s", _STR);

	    /* If the name has a path component, use only the root 	*/
	    /* for the object name, but the full path for the filename. */

	    if ((pdchar = strrchr(_STR, '/')) != NULL)
               sprintf(objectdata->name, "%s", pdchar + 1);
            else
	       sprintf(objectdata->name, "%s", _STR);

	    renamepage(areastruct.page);
	    printname(objectdata);
            Wprintf("Starting new drawing");
         }
         else {
	    sprintf(temp, "Can't find file %s, won't overwrite current page",
			_STR);
            Wprintf(temp);
         }
	 return;
      }
   }

   version = 1.0;
   multipage = 1;
   pagesize.x = 612;
   pagesize.y = 792;
   for(;;) {
      if (fgets(temp, 149, ps) == NULL) {
	 Wprintf("Error: EOF in or before prolog.");
	 return;
      }
      sscanf(temp, "%c%29s", &percentc, keyword);
      for (pdchar = keyword; isspace(*pdchar); pdchar++);
      if (percentc == '%') {
	 if (!strcmp(pdchar, "XCircuit")) break;
	 if (!strcmp(pdchar, "XCircuitLib")) {
	    /* version control in libraries is post version 1.9 */
	    if (version == 1.0) version = 1.9;
	    break;
	 }
	 if (!strcmp(pdchar, "%Page:")) break;
	 if (strstr(pdchar, "PS-Adobe") != NULL)
	    temppmode = (strstr(temp, "EPSF") != NULL) ? 0 : 1;
         else if (!strcmp(pdchar, "Version:"))
	    sscanf(temp, "%*c%*s %f", &version);
	 else if (!strcmp(pdchar, "%Pages:")) {
	    pdchar = temp;
	    for (; !isspace(*pdchar); pdchar++);
	    for (; isspace(*pdchar); pdchar++);
	    multipage = atoi(pdchar);
	 }
	 else if (!strcmp(pdchar, "%BoundingBox:")) {
	    sscanf(temp, "%*s %hd %hd %hd %hd", &offx, &offy,
		&(pagesize.x), &(pagesize.y));
	    pagesize.x += offx;
	    pagesize.y += offy;
	 }
      }
#ifdef LGF
      else if (percentc == '-' && !strcmp(keyword, "5")) {
	 fclose(ps);
	 loadlgf(mode);
	 return;
      }
#endif
   }

   /* Look for old-style files (no %%Page; maximum one page in file) */

   if (!strcmp(pdchar, "XCircuit")) {
      int pch;

      do {
	 pch = getc(ps);
      } while (pch == '\n');
      ungetc(pch, ps);
      if (pch == '%') fgets(temp, 149, ps);
   }

   for (page = 0; page < multipage; page++) {
      sprintf(pagestr, "%d", page + 1);

      /* read out-of-page library definitions */

      if (strstr(temp, "%%Page:") == NULL && strstr(temp, "offsets") == NULL)
	 objectread(ps, objectdata, 0, 0, USERLIB, temp, DEFAULTCOLOR);

      if (strstr(temp, "%%Page:") != NULL)
	 sscanf(temp + 8, "%s", pagestr);

      /* go to new page if necessary */

      if (page > 0) {

	 /* find next undefined page */

	 while(areastruct.page < xobjs.pages &&
		xobjs.pagelist[areastruct.page]->pageobj != NULL) areastruct.page++;
	 changepage(areastruct.page);
      }

      /* If this file was a library file then there is no page to load */

      if (strstr(temp, "EndLib") != NULL) {
   	 composelib(USERLIB);
	 Wprintf("Loaded library.");
	 return;
      }

      /* good so far;  let's clear out the old data structure */
      /* and set the filename and pagename */

      if (mode == 0) {
	 char tpstr[6], *rootptr;

         reset(objectdata, NORMAL);
         topreset();
	 xobjs.pagelist[areastruct.page]->pmode = temppmode;
	 if (temppmode == 1) {
	    xobjs.pagelist[areastruct.page]->pagesize.x = pagesize.x;
	    xobjs.pagelist[areastruct.page]->pagesize.y = pagesize.y;
	 }

	 /* set filename and page title.				     */
         /* for multiple-page files, filename is not linked to the page name */
	 /* but should point to a single string containing the file name and */
	 /* malloc'd from the first page of the set.			     */
	 /* for single-page files, filename is normally linked to the page   */
	 /* name unless the page label has been made different.		     */

         /* (delete the filename suffix if it is ".ps") */

	 if (page == 0) {
	    if (multipage > 1) oldpage = areastruct.page;
            xobjs.pagelist[areastruct.page]->filename = (char *)realloc(
		 xobjs.pagelist[areastruct.page]->filename, (strlen(inname) + 1)
		 * sizeof(char));
            sprintf(xobjs.pagelist[areastruct.page]->filename, "%s", inname);
            if ((pdchar = strchr(xobjs.pagelist[areastruct.page]->filename, '.')) != NULL) 
	       if (!strcmp(pdchar + 1, "ps")) *pdchar = '\0';
	 }
	 else {
	    xobjs.pagelist[areastruct.page]->filename = xobjs.pagelist[oldpage]->filename;
	 }

	 rootptr = strrchr(xobjs.pagelist[areastruct.page]->filename, '/');
	 if (rootptr == NULL) rootptr = xobjs.pagelist[areastruct.page]->filename;
	 else rootptr++;

	 sprintf(tpstr, "%d", page + 1);
	 if (!strcmp(pagestr, tpstr)) {
	    if (multipage > 1)
	       sprintf(objectdata->name, "%.72s:%s", rootptr, tpstr);
	    else
	       sprintf(objectdata->name, "%.79s", rootptr);
	 }
	 else
	    sprintf(objectdata->name, "%.79s", pagestr);

         renamepage(areastruct.page);
      }

      /* read to the "scale" line, picking up inch/cm type, drawing */
      /* scale, and grid/snapspace along the way		    */

      for(;;) {
         if (strstr(temp, "offsets") != NULL) {
            sscanf(temp, "%c %hd %hd %*s", &percentc, &offx, &offy);
            if(percentc != '%') {
               Wprintf("Something wrong in offsets line.");
               offx = offy = 0;
            }
	 }

	 if (strstr(temp, "drawingscale") != NULL)
	    sscanf(temp, "%*c %hd:%hd %*s",
		&xobjs.pagelist[areastruct.page]->drawingscale.x,
		&xobjs.pagelist[areastruct.page]->drawingscale.y);
#ifdef SCHEMA
	 else if (strstr(temp, "is_symbol") != NULL) {
	    sscanf(temp, "%*c %49s", teststr);
	    checkschem(objectdata, teststr);
	 }
#endif
	 else if (strstr(temp, "gridspace"))
	    sscanf(temp, "%*c %f %f %*s", &xobjs.pagelist[areastruct.page]->gridspace,
		 &xobjs.pagelist[areastruct.page]->snapspace);
         else if (strstr(temp, "scale") != NULL || strstr(temp, "rotate") != NULL) {
            /* rotation (landscape mode) is optional; parse accordingly */

            sscanf(temp, "%f %s", &tmpfl, teststr);
            if (strstr(teststr, "scale") != NULL) {
	       setgridtype(teststr);
               xobjs.pagelist[areastruct.page]->outscale = tmpfl;
            }
            else if (!strcmp(teststr, "rotate")) {
               xobjs.pagelist[areastruct.page]->orient = (short)tmpfl;
               fgets(temp, 149, ps);
               sscanf(temp, "%f %s", &tmpfl, teststr2);
	       if (strstr(teststr2, "scale") != NULL) {
	          setgridtype(teststr2);
	          xobjs.pagelist[areastruct.page]->outscale = tmpfl;
	       }
	       else {
	          sscanf(temp, "%*f %*f %s", teststr2);
	          if (!strcmp(teststr2, "scale"))
	             xobjs.pagelist[areastruct.page]->outscale = tmpfl /
		          getpsscale(1.0, areastruct.page);
	          else {
	             Wprintf("Error in scale/rotate constructs.");
	             return;
	          }
	       }
            }
            else {     /* old style scale? */
               sscanf(temp, "%*f %*s %s", teststr2);
	       if ((teststr2 != NULL) && (!strcmp(teststr2, "scale")))
                  xobjs.pagelist[areastruct.page]->outscale = tmpfl /
		        getpsscale(1.0, areastruct.page);
	       else {
                  Wprintf("Error in scale/rotate constructs.");
                  return;
	       }
            }
	 }
         else if (strstr(temp, "setlinewidth") != NULL) {
            sscanf(temp, "%f %*s", &xobjs.pagelist[areastruct.page]->wirewidth);
            xobjs.pagelist[areastruct.page]->wirewidth /= 1.3;
	    break;
	 }
	 else if (strstr(temp, "insertion") != NULL) {
	    /* read in an included background image */
	    readbackground(ps);
	 }

	 if (fgets(temp, 149, ps) == NULL) {
            Wprintf("Error: Problems encountered in page header.");
            return;
         }
      }

      objectread(ps, objectdata, offx, offy, LIBRARY, temp, DEFAULTCOLOR);

      /* skip to next page boundary or file trailer */

      if (strstr(temp, "showpage") != NULL && multipage != 1) {
	 int pch;
         do {
	    pch = getc(ps);
         } while (pch == '\n');
	 ungetc(pch, ps);
         if (pch == '%') fgets(temp, 149, ps);
      }

      /* set object position to fit to window separately for each page */
      centerview(objectdata);
   }

   sprintf(_STR, "Loaded file: %s (%d page%s)", inname, multipage,
	(multipage > 1 ? "s" : ""));
   Wprintf(_STR);

   composelib(USERLIB);
   composelib(PAGELIB);

   if (version > PROG_VERSION) {
      sprintf(_STR, "WARNING: file %s is version %2.1f vs. executable version %2.1f",
		inname, version, PROG_VERSION);
      Wprintf(_STR);
   }

   version = PROG_VERSION;
}

/*------------------------------------------------------*/
/* Object name comparison:  True if names are equal,	*/
/* not counting leading underscores.			*/
/*------------------------------------------------------*/

int objnamecmp(char *name1, char *name2)
{
   char *n1ptr = name1;
   char *n2ptr = name2;

   while (*n1ptr == '_') n1ptr++;
   while (*n2ptr == '_') n2ptr++;

   return (strcmp(n1ptr, n2ptr));
}

/*------------------------------------------------------*/
/* Find matching parenthesis, bracket, or brace		*/
/* Don't count when backslash character '\' is in front */
/*------------------------------------------------------*/

u_char *find_match(u_char *fstring)
{
   int count = 1;
   u_char *search = fstring; 
   u_char source = *fstring;
   u_char target;

   switch(source) {
      case '(':  target = ')'; break;
      case '[':  target = ']'; break;
      case '{':  target = '}'; break;
      default:   target = source;
   }

   while (*++search != '\0') {
      if (*search == source && *(search - 1) != '\\') count++;
      else if (*search == target && *(search - 1) != '\\') count--;
      if (count == 0) break;
   }
   return search;
}

/*----------------------------------------------------------------------*/
/* Remove unnecessary font change information from a label		*/
/*----------------------------------------------------------------------*/

void cleanuplabel(stringpart **strhead)
{
   stringpart *curpart = *strhead;
   int oldfont, curfont;

   oldfont = curfont = -1;

   while (curpart != NULL) {
      switch (curpart->type) {
	 case FONT_NAME:
	    if (curfont == curpart->data.font) {
	       /* Font change is redundant:  remove it */
	       curpart = deletestring(curpart, strhead, NORMINST);
	    }
	    else {
	       curfont = curpart->data.font;
	    }
	    break;

	 case FONT_SCALE:
	    /* Old style font scale is always written absolute, not relative.	*/
	    /* Changes in scale were not allowed, so just get rid of them.	*/
	    if (version < 2.25)
	       curpart = deletestring(curpart, strhead, NORMINST);
	    break;

	 /* A font change may occur inside a parameter, so any font	*/
	 /* declaration after a parameter must be considered to be	*/
	 /* intentional.						*/

	 case PARAM_END:
	    curfont = oldfont = -1;
	    break;

	 case NORMALSCRIPT: case RETURN:
	    if (oldfont != -1) {
	       curfont = oldfont;
	       oldfont = -1;
	    }
	    break;

	 case SUBSCRIPT: case SUPERSCRIPT:
	    if (oldfont == -1)
	       oldfont = curfont;
	    break;
      }
      if (curpart != NULL)
	 curpart = curpart->nextpart;
   }
}

/*----------------------------------------------------------------------*/
/* Read label segments							*/
/*----------------------------------------------------------------------*/

void readlabel(objectptr localdata, char *lineptr, stringpart **strhead)
{
   Boolean fline = False;
   char *sptr;
   short j;
   int paramno;
   char *endptr, *segptr = lineptr;
   stringpart *newpart;

   while (*segptr != '\0') {	/* Look through all segments */

      while (isspace(*segptr) && (*segptr != '\0')) segptr++;

      if (*segptr == '(' || *segptr == '{') {
         endptr = find_match(segptr);
         *endptr++ = '\0';
	 /* null string (e.g., null parameter substitution) */
	 if ((*segptr == '(') && (*(segptr + 1) == '\0')) {
	    segptr = endptr;
	    continue;
	 }
      }
      else if (*segptr == '\0' || *segptr == '}') break;

      makesegment(strhead, *strhead);
      newpart = *strhead;

      /* Embedded command is in braces: {} */

      if (*segptr == '{') {	

         /* Find the command for this PostScript procedure */
	 char *cmdptr = endptr - 2;
         while (isspace(*cmdptr)) cmdptr--;
         while (!isspace(*cmdptr) && (cmdptr > segptr)) cmdptr--;
	 cmdptr++;
	 segptr++;

         if (!strncmp(cmdptr, "Ss", 2))
	    newpart->type = SUPERSCRIPT;
         else if (!strncmp(cmdptr, "ss", 2))
            newpart->type = SUBSCRIPT;
         else if (!strncmp(cmdptr, "ns", 2))
            newpart->type = NORMALSCRIPT;
         else if (!strncmp(cmdptr, "hS", 2))
            newpart->type = HALFSPACE;
         else if (!strncmp(cmdptr, "qS", 2))
            newpart->type = QTRSPACE;
         else if (!strncmp(cmdptr, "CR", 2))
	    newpart->type = RETURN;
         else if (!strcmp(cmdptr, "Ts")) /* "Tab set" command */
	    newpart->type = TABSTOP;
         else if (!strcmp(cmdptr, "Tf")) /* "Tab forward" command */
	    newpart->type = TABFORWARD;
         else if (!strcmp(cmdptr, "Tb")) /* "Tab backward" command */
	    newpart->type = TABBACKWARD;
         else if (!strncmp(cmdptr, "ol", 2)) {
            newpart->type = OVERLINE;
            fline = True;
         }
         else if (!strncmp(cmdptr, "ul", 2)) {
            newpart->type = UNDERLINE;
            fline = True;
         }
         else if (cmdptr == segptr) {   /* cancel over- or underline */
            newpart->type = NOLINE;
            fline = False;
         }
	 /* To-do:  replace old-style backspace with tab stop */
         else if (!strcmp(cmdptr, "bs")) {
	    Wprintf("Warning:  Obsolete backspace command ommitted in text");
         }
         else if (!strcmp(cmdptr, "Kn")) {  /* "Kern" command */
	    int kx, ky;
	    sscanf(segptr, "%d %d", &kx, &ky);  
	    newpart->type = KERN;
	    newpart->data.kern[0] = kx;
	    newpart->data.kern[1] = ky;
         }
         else if (!strcmp(cmdptr, "scb")) {  /* change color command */
	    float cr, cg, cb;
	    int cval, cindex;
	    sscanf(segptr, "%f %f %f", &cr, &cg, &cb);  
            newpart->type = FONT_COLOR;
	    cval = rgb_alloccolor((int)(cr * 65535), (int)(cg * 65535),
		   (int)(cb * 65535));
	    for (cindex = 0; cindex < number_colors; cindex++)
               if (colorlist[cindex].color.pixel == cval)
                  break;
	    if (cindex == number_colors) {
	       Wprintf("Error:  Cannot allocate another color");
	       cindex = DEFAULTCOLOR;
	    }
	    newpart->data.color = cindex;
         }
         else if (!strcmp(cmdptr, "sce")) {  /* revert to default color */
            newpart->type = FONT_COLOR;
	    newpart->data.color = DEFAULTCOLOR;
         }
         else if (!strcmp(cmdptr, "cf")) {  /* change font or scale command */
	    char *nextptr, *newptr = segptr;

	    /* Set newptr to the fontname and nextptr to the next token. */
	    while (*newptr != '/' && *newptr != '\0') newptr++;
	    if (*newptr++ == '\0') {
	       Wprintf("Error:  Bad change-font command");
	       return;
	    }
	    for (nextptr = newptr; !isspace(*nextptr); nextptr++);
	    *(nextptr++) = '\0';
	    while (isspace(*nextptr)) nextptr++;

            for (j = 0; j < fontcount; j++)
               if (!strcmp(newptr, fonts[j].psname))
	          break;

            if (j == fontcount) 		/* this is a non-loaded font */
               loadfontfile(newptr);

            if (isdigit(*nextptr)) { /* second form of "cf" command---includes scale */
	       float locscale;
	       sscanf(nextptr, "%f", &locscale);
	       if (locscale != 1.0) {
	          newpart->type = FONT_SCALE;
	          newpart->data.scale = locscale;
	          makesegment(strhead, *strhead);
	          newpart = *strhead;
	       }
            }
	    newpart->type = FONT_NAME;
	    newpart->data.font = j;
         }
         else {   /* This exec isn't a known label function */
	    Wprintf("Error:  unknown substring function");
	    newpart->type = NOLINE;	/* This is fairly benign */
         }
      }

      /* Text substring is in parentheses: () */

      else if (*segptr == '(') {
	 u_char *tmpptr;
         if (fline == True) {
	    newpart->type = NOLINE;
            makesegment(strhead, *strhead);
	    newpart = *strhead;
	    fline = False;
         }
         newpart->type = TEXT_STRING;
         newpart->data.string = (u_char *)malloc(1 + strlen(++segptr));

         /* Copy string, translating octal codes into 8-bit characters */

         tmpptr = newpart->data.string;
         for (sptr = segptr; *sptr != '\0'; sptr++) {
            if (*sptr == '\\') {
               sptr++;
	       if (*sptr >= '0' && *sptr < '8') {
	          int tmpdig;
	          sscanf(sptr, "%3o", &tmpdig);
	          *tmpptr++ = (u_char)tmpdig;
	          sptr += 2;
	       }
               else *tmpptr++ = (u_char) *sptr;
	    }
            else *tmpptr++ = (u_char) *sptr;
	 }
 	 *tmpptr = '\0';
      }

      /* Parameterized substrings are denoted v1, v2, v3,... 	*/
      /* The parameter (default and/or substitution value) is 	*/
      /* assumed to exist.					*/

      else {

         if (sscanf(segptr + 1, "%d", &paramno) == 1) {
	    newpart->type = PARAM_START;
	    newpart->data.paramno = paramno - 1;

            /* printf("Parameter %d called from object %s\n", paramno,	*/
	    /*		localdata->name);				*/
         }
	 endptr = segptr + 1;
	 while (!isspace(*endptr)) endptr++;
      }
      segptr = endptr;
   }
}

/*--------------------------------------------------------------*/
/* Read a parameter point					*/
/*--------------------------------------------------------------*/

void varscan(objectptr localdata, char *lineptr, short *dpoint)
{
   oparamptr ops;
   int paramno;
   char *varchr;

   if (sscanf(lineptr, "%hd", dpoint) != 1) {
      if ((varchr = strchr(lineptr + 1, 'v')) != NULL) {
         if (sscanf(varchr + 1, "%d", &paramno) == 1) {
            ops = localdata->params[paramno - 1];

	    /* Currently there is no procedure for integer parameters */
	 }
	 else 
	    Wprintf("Error in reading parameter point substitution");
      }
      else
	 Wprintf("Error in point specification");
   }
}

/*--------------------------------------------------------------*/
/* Continuation Line --- add memory to "buffer" as necessary.	*/
/* Add a space character to the current text in "buffer" and	*/
/* return a pointer to the new end-of-text.			*/
/*--------------------------------------------------------------*/

char *continueline(char **buffer)
{
   char *lineptr;
   int bufsize;

   for (lineptr = *buffer; (*lineptr != '\n') && (*lineptr != '\0'); lineptr++);
   if (*lineptr == '\n') *lineptr++ = ' ';

   bufsize = (int)(lineptr - (*buffer)) + 256;
   *buffer = (char *)realloc((*buffer), bufsize * sizeof(char));

   return ((*buffer) + (bufsize - 256));
}

/*--------------------------------------------------------------*/
/* Read an object (page) from a file into xcircuit		*/
/*--------------------------------------------------------------*/

Boolean objectread(FILE *ps, objectptr localdata, short offx, short offy,
	short mode, char *retstr, int ccolor)
{
   char *temp, *buffer, keyword[80]; 
   short tmpfont = -1;
   float tmpscale = 0.0;
   objectptr	*libobj;
   int curcolor = ccolor;
   int i, j, k;

   /* path-handling variables */
   pathptr *newpath;
   XPoint startpoint;

   keyword[0] = '\0';

   buffer = (char *)malloc(256 * sizeof(char));
   temp = buffer;

   for(;;) {
      char *lineptr, *keyptr;

      if (fgets(temp, 255, ps) == NULL) {
	 if (strcmp(keyword, "restore")) {
            Wprintf("Error: end of file.");
	    *retstr = '\0';
	 }
	 break;
      }
      temp = buffer;

      /* because PostScript is a stack language, we will scan from the end */
      for (lineptr = buffer; (*lineptr != '\n') && (*lineptr != '\0'); lineptr++);
      if (lineptr != buffer) {  /* ignore any blank lines */
         for (keyptr = lineptr - 1; isspace(*keyptr) && keyptr != buffer; keyptr--);
         for (; !isspace(*keyptr) && keyptr != buffer; keyptr--);
         sscanf(keyptr, "%79s", keyword);

         if (!strcmp(keyword, "showpage")) {
            strncpy(retstr, buffer, 150);
            retstr[149] = '\0';
	    free(buffer);
	    calcbbox(localdata);
	    return False;  /* end of page */
	 }

	 /* make a color change, adding the color if necessary */

	 else if (!strcmp(keyword, "scb")) {
	    float red, green, blue;
	    sscanf(buffer, "%f %f %f", &red, &green, &blue);
	    curcolor = rgb_alloccolor((int)(red * 65535), (int)(green * 65535),
		(int)(blue * 65535)); 
	    addnewcolorentry(curcolor);
	 }

	 /* end the color change, returning to default */

	 else if (!strcmp(keyword, "sce")) curcolor = ccolor;

	 /* begin a path constructor */

	 else if (!strcmp(keyword, "beginpath")) {
	    sscanf(buffer, "%hd %hd", &startpoint.x, &startpoint.y);
	    startpoint.x -= offx; startpoint.y -= offy;
	    NEW_PATH(newpath, localdata);
	    (*newpath)->plist = (genericptr *)malloc(sizeof(genericptr));
	    (*newpath)->parts = 0;
	    (*newpath)->color = curcolor;
	 }

	 /* end the path constructor */

	 else if (!strcmp(keyword, "endpath")) {
	    sscanf(buffer, "%hd %f", &(*newpath)->style, &(*newpath)->width);
	    if ((*newpath)->parts > 0) {
	       localdata->parts++;
	    }
	    else {	/* in case of an empty path */
	       free((*newpath)->plist);
	       free(*newpath);
	    }
	    newpath = NULL;
	 }

	 /* read path parts */

	 else if (!strcmp(keyword, "polyc")) {
	    polyptr *newpoly;
	    pointlist newpoints;
	    short tmpnum;

	    NEW_POLY(newpoly, (*newpath));
	    (*newpath)->parts++;

	    for (--keyptr; *keyptr == ' '; keyptr--);
	    for (; *keyptr != ' '; keyptr--);
	    sscanf(keyptr, "%hd", &tmpnum);
	    (*newpoly)->number = tmpnum + 1;
	    (*newpoly)->width = 1.0;
	    (*newpoly)->style = UNCLOSED;
	    (*newpoly)->color = curcolor;

            (*newpoly)->points = (pointlist) malloc((*newpoly)->number * 
		   sizeof(XPoint));

	    lineptr = buffer;
            for (newpoints = (*newpoly)->points + (*newpoly)->number - 1;
		   newpoints > (*newpoly)->points; newpoints--) {
	       sscanf(lineptr, "%hd", &newpoints->x); 
	       newpoints->x -= offx;
	       for (; *lineptr == ' '; lineptr++); 
	       for (; *lineptr != ' '; lineptr++); 
	       sscanf(lineptr, "%hd", &newpoints->y);
	       newpoints->y -= offy;
	       for (; *lineptr == ' '; lineptr++); 
	       for (; *lineptr != ' '; lineptr++); 
	    }
	    newpoints->x = startpoint.x;
	    newpoints->y = startpoint.y;
	    startpoint.x = (newpoints + (*newpoly)->number - 1)->x;
	    startpoint.y = (newpoints + (*newpoly)->number - 1)->y;
	 }

	 else if (!strcmp(keyword, "arc") || !strcmp(keyword, "arcn")) {
	    arcptr *newarc;
	    NEW_ARC(newarc, (*newpath));
	    (*newpath)->parts++;
	    sscanf(buffer, "%hd %hd %hd %f %f", &(*newarc)->position.x,
		  &(*newarc)->position.y, &(*newarc)->radius,
		  &(*newarc)->angle1, &(*newarc)->angle2);
	    (*newarc)->position.x -= offx; (*newarc)->position.y -= offy;
	    (*newarc)->yaxis = (*newarc)->radius;
	    (*newarc)->width = 1.0;
	    (*newarc)->style = UNCLOSED;
	    (*newarc)->color = curcolor;
	    if (!strcmp(keyword, "arcn")) {
	       float tmpang = (*newarc)->angle1;
	       (*newarc)->radius = -((*newarc)->radius);
	       (*newarc)->angle1 = (*newarc)->angle2;
	       (*newarc)->angle2 = tmpang;
	    }
		
	    calcarc(*newarc);
	    startpoint.x = (short)(*newarc)->points[(*newarc)->number - 1].x;
	    startpoint.y = (short)(*newarc)->points[(*newarc)->number - 1].y;
	 }

	 else if (!strcmp(keyword, "pellip") || !strcmp(keyword, "nellip")) {
	    arcptr *newarc;
	    NEW_ARC(newarc, (*newpath));
	    (*newpath)->parts++;
	    sscanf(buffer, "%hd %hd %hd %hd %f %f", &(*newarc)->position.x,
		 &(*newarc)->position.y, &(*newarc)->radius, &(*newarc)->yaxis,
		 &(*newarc)->angle1, &(*newarc)->angle2);
	    (*newarc)->position.x -= offx; (*newarc)->position.y -= offy;
	    (*newarc)->width = 1.0;
	    (*newarc)->style = UNCLOSED;
	    (*newarc)->color = curcolor;
	    if (!strcmp(keyword, "nellip")) {
	       float tmpang = (*newarc)->angle1;
	       (*newarc)->radius = -((*newarc)->radius);
	       (*newarc)->angle1 = (*newarc)->angle2;
	       (*newarc)->angle2 = tmpang;
		
	    }
	    calcarc(*newarc);
	    startpoint.x = (short)(*newarc)->points[(*newarc)->number - 1].x;
	    startpoint.y = (short)(*newarc)->points[(*newarc)->number - 1].y;
	 }

	 else if (!strcmp(keyword, "curveto")) {
	    splineptr *newspline;
	    NEW_SPLINE(newspline, (*newpath));
	    (*newpath)->parts++;
            sscanf(buffer, "%hd %hd %hd %hd %hd %hd", 
		  &(*newspline)->ctrl[1].x, &(*newspline)->ctrl[1].y,
		  &(*newspline)->ctrl[2].x, &(*newspline)->ctrl[2].y,
		  &(*newspline)->ctrl[3].x, &(*newspline)->ctrl[3].y);
	    (*newspline)->ctrl[0].x = startpoint.x;
	    (*newspline)->ctrl[0].y = startpoint.y;
	    (*newspline)->ctrl[1].x -= offx; (*newspline)->ctrl[2].x -= offx;
	    (*newspline)->ctrl[3].x -= offx;
	    (*newspline)->ctrl[1].y -= offy; (*newspline)->ctrl[2].y -= offy;
	    (*newspline)->ctrl[3].y -= offy;
	    (*newspline)->width = 1.0;
	    (*newspline)->style = UNCLOSED;
	    (*newspline)->color = curcolor;
	    calcspline(*newspline);
	    startpoint.x = (*newspline)->ctrl[3].x;
	    startpoint.y = (*newspline)->ctrl[3].y;
	 }

         /* read arcs */

         else if (!strcmp(keyword, "xcarc")) {
	    arcptr *newarc;
	 
	    NEW_ARC(newarc, localdata);

	    /* backward compatibility */
	    if (version < 1.5) 
	       sscanf(buffer, "%hd %hd %hd %f %f %f %hd", &(*newarc)->position.x,
	          &(*newarc)->position.y, &(*newarc)->radius, &(*newarc)->angle1,
	          &(*newarc)->angle2, &(*newarc)->width, &(*newarc)->style);
	    else
	       sscanf(buffer, "%hd %f %hd %hd %hd %f %f", &(*newarc)->style,
		  &(*newarc)->width, &(*newarc)->position.x, &(*newarc)->position.y,
		  &(*newarc)->radius, &(*newarc)->angle1, &(*newarc)->angle2);

	    (*newarc)->position.x -= offx; (*newarc)->position.y -= offy;
	    (*newarc)->yaxis = (*newarc)->radius;
	    (*newarc)->color = curcolor;
	    calcarc(*newarc);
	    localdata->parts++;
         }

	 /* read ellipses */

         else if (!strcmp(keyword, "ellipse")) {
	    arcptr *newarc;
	 
	    NEW_ARC(newarc, localdata);

	    sscanf(buffer, "%hd %f %hd %hd %hd %hd %f %f", &(*newarc)->style,
		 &(*newarc)->width, &(*newarc)->position.x, &(*newarc)->position.y,
		 &(*newarc)->radius, &(*newarc)->yaxis, &(*newarc)->angle1,
		 &(*newarc)->angle2);

	    (*newarc)->position.x -= offx; (*newarc)->position.y -= offy;
	    (*newarc)->color = curcolor;
	    calcarc(*newarc);
	    localdata->parts++;
         }

         /* read polygons */
	 /* (and wires---backward compatibility for v1.5 and earlier) */

         else if (!strcmp(keyword, "polygon") || !strcmp(keyword, "wire")) {
	    polyptr *newpoly;
	    pointlist newpoints;

	    NEW_POLY(newpoly, localdata);
	    lineptr = buffer;

	    if (!strcmp(keyword, "wire")) {
	       (*newpoly)->number = 2;
	       (*newpoly)->width = 1.0;
	       (*newpoly)->style = UNCLOSED;
	    }
	    else {
	       /* backward compatibility */
	       if (version < 1.5) {
	          for (--keyptr; *keyptr == ' '; keyptr--);
	          for (; *keyptr != ' '; keyptr--);
	          sscanf(keyptr, "%hd", &(*newpoly)->style);
	          for (--keyptr; *keyptr == ' '; keyptr--);
	          for (; *keyptr != ' '; keyptr--);
	          sscanf(keyptr, "%f", &(*newpoly)->width);
	       }
	       for (--keyptr; *keyptr == ' '; keyptr--);
	       for (; *keyptr != ' '; keyptr--);
	       sscanf(keyptr, "%hd", &(*newpoly)->number);

	       if (version >= 1.5) {
	          sscanf(lineptr, "%hd", &(*newpoly)->style);
	          for (; *lineptr == ' '; lineptr++);
	          for (; *lineptr != ' '; lineptr++); 
	          sscanf(lineptr, "%f", &(*newpoly)->width);
	          for (; *lineptr == ' '; lineptr++);
	          for (; *lineptr != ' '; lineptr++); 
	       }
	    }

	    if ((*newpoly)->style & BBOX)
	       (*newpoly)->color = BBOXCOLOR;
	    else
	       (*newpoly)->color = curcolor;
            (*newpoly)->points = (pointlist) malloc((*newpoly)->number * 
		   sizeof(XPoint));

            for (newpoints = (*newpoly)->points; newpoints < (*newpoly)->points
		+ (*newpoly)->number; newpoints++) {
	       varscan(localdata, lineptr, &newpoints->x);
	       newpoints->x -= offx;
	       for (; *lineptr == ' '; lineptr++); 
	       for (; *lineptr != ' '; lineptr++); 
	       varscan(localdata, lineptr, &newpoints->y);
	       newpoints->y -= offy;
	       for (; *lineptr == ' '; lineptr++); 
	       for (; *lineptr != ' '; lineptr++); 
	    }
	    localdata->parts++;
         }

	 /* read spline curves */

         else if (!strcmp(keyword, "spline")) {
            splineptr *newspline;

	    NEW_SPLINE(newspline, localdata);

	    /* backward compatibility */
	    if (version < 1.5) {
               sscanf(buffer, "%f %hd %hd %hd %hd %hd %hd %hd %hd %hd", 
	          &(*newspline)->width, &(*newspline)->ctrl[1].x,
	          &(*newspline)->ctrl[1].y, &(*newspline)->ctrl[2].x,
	          &(*newspline)->ctrl[2].y, &(*newspline)->ctrl[3].x,
	          &(*newspline)->ctrl[3].y, &(*newspline)->ctrl[0].x,
	          &(*newspline)->ctrl[0].y, &(*newspline)->style);
	    }
	    else {
               sscanf(buffer, "%hd %f %hd %hd %hd %hd %hd %hd %hd %hd", 
		  &(*newspline)->style, &(*newspline)->width,
		  &(*newspline)->ctrl[1].x, &(*newspline)->ctrl[1].y,
		  &(*newspline)->ctrl[2].x, &(*newspline)->ctrl[2].y,
		  &(*newspline)->ctrl[3].x, &(*newspline)->ctrl[3].y,
		  &(*newspline)->ctrl[0].x, &(*newspline)->ctrl[0].y);
	    }

	    (*newspline)->ctrl[1].x -= offx; (*newspline)->ctrl[2].x -= offx;
	    (*newspline)->ctrl[0].x -= offx;
	    (*newspline)->ctrl[3].x -= offx;
	    (*newspline)->ctrl[1].y -= offy; (*newspline)->ctrl[2].y -= offy;
	    (*newspline)->ctrl[3].y -= offy;
	    (*newspline)->ctrl[0].y -= offy;
	    (*newspline)->color = curcolor;
            localdata->parts++;
	    calcspline(*newspline);
         }

         /* read labels */

         else if (!strcmp(keyword, "fontset")) { 	/* old style */
            char tmpstring[100];
            int i;
            sscanf(buffer, "%f %*c%s", &tmpscale, tmpstring);
            for (i = 0; i < fontcount; i++)
               if (!strcmp(tmpstring, fonts[i].psname)) {
		  tmpfont = i;
		  break;
	       }
	    if (i == fontcount) i = 0;	/* Why bother with anything fancy? */
         }

         else if (!strcmp(keyword, "label") || !strcmp(keyword, "pinlabel")
		|| !strcmp(keyword, "pinglobal") || !strcmp(keyword, "infolabel")) {

	    labelptr *newlabel;
	    stringpart *firstscale, *firstfont;

	    NEW_LABEL(newlabel, localdata);

	    /* scan backwards to get the number of substrings */
	    lineptr = keyptr - 1;
	    for (i = 0; i < ((version < 2.25) ? 5 : 6); i++) {
	       for (; *lineptr == ' '; lineptr--);
	       for (; *lineptr != ' '; lineptr--);
	    }
	    if ((strchr(lineptr, '.') != NULL) && (version < 2.25)) {
	       fprintf(stderr, "Error:  File version claims to be %2.1f,"
			" but has version %2.1f labels\n", version, PROG_VERSION);
	       fprintf(stderr, "Attempting to resolve problem by updating version.\n");
	       version = PROG_VERSION;
	       for (; *lineptr == ' '; lineptr--);
	       for (; *lineptr != ' '; lineptr--);
	    }
	    /* no. segments is ignored---may be a derived quantity, anyway */
	    if (version < 2.25) {
	       sscanf(lineptr, "%*s %hd %hd %hd %hd", &(*newlabel)->justify,
		   &(*newlabel)->rotation, &(*newlabel)->position.x,
		   &(*newlabel)->position.y);
	    }
	    else {
	       sscanf(lineptr, "%*s %hd %hd %f %hd %hd", &(*newlabel)->justify,
		   &(*newlabel)->rotation, &(*newlabel)->scale,
		   &(*newlabel)->position.x, &(*newlabel)->position.y);
	    }
	    (*newlabel)->position.x -= offx; (*newlabel)->position.y -= offy;
	    if ((*newlabel)->rotation > 0)
	       (*newlabel)->rotation = 360 - (*newlabel)->rotation;
	    (*newlabel)->rotation /= ROT_UNIT;
	    (*newlabel)->rotation++;
	    (*newlabel)->color = curcolor;
	    (*newlabel)->string = NULL;
	    *lineptr = '\0';

#ifdef SCHEMA
	    (*newlabel)->pin = False;
	    if (strcmp(keyword, "label")) {	/* all the schematic types */
	       /* enable schematic capture if it is not already on. */
	       if (!areastruct.schemon) doxschema(NULL, NULL, NULL);
	       if (!strcmp(keyword, "pinlabel"))
		  (*newlabel)->pin = LOCAL;
	       else if (!strcmp(keyword, "pinglobal"))
		  (*newlabel)->pin = GLOBAL;
	       else if (!strcmp(keyword, "infolabel")) {
		  localdata->schemtype = FUNDAMENTAL;
		  (*newlabel)->pin = INFO;
		  if (curcolor == DEFAULTCOLOR)
		     (*newlabel)->color = AUXCOLOR;
	       }
	    }
#endif
	    lineptr = buffer;	/* back to beginning of string */
	    if (!strncmp(lineptr, "mark", 4)) lineptr += 4;

	    readlabel(localdata, lineptr, &(*newlabel)->string);

	    if (version < 2.25) {
	       /* Switch 1st scale designator to overall font scale */

	       firstscale = (*newlabel)->string->nextpart;
	       if (firstscale->type != FONT_SCALE) {
	          if (tmpscale != 0.0)
	             (*newlabel)->scale = 0.0;
	          else
	             (*newlabel)->scale = 1.0;
	       }
	       else {
	          (*newlabel)->scale = firstscale->data.scale;
	          deletestring(firstscale, &((*newlabel)->string), NORMINST);
	       }
	    }

	    firstfont = (*newlabel)->string;
	    if (firstfont->type != FONT_NAME) {
	       if (tmpfont == -1) {
	          fprintf(stderr, "Error:  Label with no font designator?\n");
		  tmpfont = 0;
	       }
	       firstfont = makesegment(&((*newlabel)->string), (*newlabel)->string);  
	       firstfont->type = FONT_NAME;
	       firstfont->data.font = tmpfont;
	    }
	    cleanuplabel(&(*newlabel)->string);

#ifdef SCHEMA

	    /* link string to the one that corresponds to this one */
	    /* in the symbol, if it exists.			   */

	    if (areastruct.schemon && localdata->symschem != NULL &&
		  (*newlabel)->pin && ((*newlabel)->pin != INFO)) {
	       genericptr *pgen;
	       for (pgen = localdata->symschem->plist; pgen < localdata->
			symschem->plist + localdata->symschem->parts; pgen++) {
		  if ((*pgen)->type == LABEL)
		     if (!stringcomp(TOLABEL(pgen)->string, (*newlabel)->string))
		 	break;
	       }
	       if (pgen == localdata->symschem->plist + localdata->symschem->parts)
		  Wprintf("Error:  Unattached pin");
	    }
#endif

	    localdata->parts++;
         }

	 /* read symbol-to-schematic connection */

#ifdef SCHEMA
	 else if (!strcmp(keyword, "is_schematic")) {
	    char tempstr[50];
	    for (lineptr = buffer; *lineptr == ' '; lineptr++);
	    sscanf(++lineptr, "%49s", tempstr);
	    checksym(localdata, tempstr);
	 }
#endif

         /* read bounding box */

         else if (!strcmp(keyword, "bbox")) {
	    for (lineptr = buffer; *lineptr == ' '; lineptr++);
            if (*lineptr != '%') {
	       Wprintf("Illegal bbox.");
	       free(buffer);
               *retstr = '\0';
	       return True;
	    }
#ifdef SCHEMA
	    sscanf(++lineptr, "%hd %hd %hd %hd", &localdata->lleft2.x, 
	       &localdata->lleft2.y, &localdata->width2, &localdata->height2);
	    if (mode == FONTLIB) {
	       localdata->lowerleft.x = localdata->lleft2.x;
	       localdata->lowerleft.y = localdata->lleft2.y;
	       localdata->width = localdata->width2;
	       localdata->height = localdata->height2;
	    }
#else
	    sscanf(++lineptr, "%hd %hd %hd %hd", &localdata->lowerleft.x, 
	       &localdata->lowerleft.y, &localdata->width, &localdata->height);
#endif
         }

	 /* read "hidden" attribute */

	 else if (!strcmp(keyword, "hidden")) {
	    localdata->hidden = True;
	 }

         /* read objects */

         else if (!strcmp(keyword, "{")) {  /* This is an object definition */
	    objectpairptr newdef, redef = NULL;
	    objectptr *newobject;
	    objectptr *curlib = (mode == FONTLIB) ?
		xobjs.fontlib.library : xobjs.userlibs[mode - LIBRARY].library;
	    short *libobjects = (mode == FONTLIB) ?
		&xobjs.fontlib.number : &xobjs.userlibs[mode - LIBRARY].number;

	    for (lineptr = buffer; *lineptr == ' '; lineptr++);
	    if (*lineptr++ != '/') {
	       /* This may be part of a label. . . treat as a continuation line */
	       temp = continueline(&buffer);
	       continue;
	    }
	    sscanf(lineptr, "%79s", keyword);

	    curlib = (objectptr *) realloc(curlib, (*libobjects + 1)
		      * sizeof(objectptr));
	    if (mode == FONTLIB) xobjs.fontlib.library = curlib;
	    else xobjs.userlibs[mode - LIBRARY].library = curlib;
	    /* initial 1-pointer allocations */
            newobject = curlib + (*libobjects);
	    *newobject = (objectptr) malloc(sizeof(object));
	    initmem(*newobject);

	    /* check that this object is not already in list of objects */

	    if (mode == FONTLIB) {
	       for (libobj = xobjs.fontlib.library; libobj != xobjs.fontlib.library +
	            xobjs.fontlib.number; libobj++) {
	          if (!objnamecmp(keyword, (*libobj)->name)) {  /* Is a redefinition */
		     newdef = (objectpairptr) malloc(sizeof(objectpair));
		     newdef->thisobject = *libobj;
		     newdef->nextpair = redef;
		     redef = newdef;
	          }
	       }
	    }
	    else {
	       for (i = 0; i < xobjs.numlibs; i++) {
		  for (j = 0; j < xobjs.userlibs[i].number; j++) {
		     libobj = xobjs.userlibs[i].library + j;
	             if (!objnamecmp(keyword, (*libobj)->name)) {  /* Is a redefinition */
		        newdef = (objectpairptr) malloc(sizeof(objectpair));
		        newdef->thisobject = *libobj;
		        newdef->nextpair = redef;
		        redef = newdef;
		     }
	          }
	       }
	    }

            (*libobjects)++;
	    sprintf((*newobject)->name, "%s", keyword);

#ifdef SCHEMA
	    /* initmem() initialized schemtype to SCHEMATIC;  change it. */
	    (*newobject)->schemtype = SYMBOL;
#endif

	    if (objectread(ps, *newobject, 0, 0, mode, retstr, curcolor) == True) {
               strncpy(retstr, buffer, 150);
               retstr[149] = '\0';
	       free(buffer);
	       return True;
            }

	    /* do an exhaustive comparison between this object 	*/
	    /* and any object having the same name or same name	*/
	    /* altered by appending an underscore.  If they are	*/
	    /* the same, destroy the duplicate.  If different,	*/
	    /* and names are the same, rename the original one.	*/

	    else if (redef != NULL) {
	       for (newdef = redef; newdef != NULL; newdef = newdef->nextpair) {
		  if (!strcmp((*newobject)->name, newdef->thisobject->name)) {
		     redef = redef; 	/* place to break on */
		  }
	          if (objcompare(*newobject, newdef->thisobject) == True) {
		     strcpy(newdef->thisobject->name, (*newobject)->name);
		     reset(*newobject, DESTROY);
		     (*libobjects)--;
		     break;
	          }
		  else if (!strcmp((*newobject)->name, newdef->thisobject->name)) {
		     strcpy(_STR, newdef->thisobject->name);
		     checkname(newdef->thisobject);
		  }
	       }
	       for (; (newdef = redef->nextpair); redef = newdef)
		  free(redef);
	       free(redef);
	       redef = NULL;
	    }
         }
         else if (!strcmp(keyword, "def")) {
            strncpy(retstr, buffer, 150);
            retstr[149] = '\0';
	    free (buffer);
	    if (mode != FONTLIB) {
	       calcbboxvalues(localdata, (u_char)1, (genericptr *)NULL);
	       centerview(localdata);
	    }
	    return False; /* end of object def or end of object library */
	 }

	 else if (!strcmp(keyword, "loadfontencoding")) {
	    /* Deprecated, but retained for backward compatibility. */
	    /* Load from script, .xcircuitrc, or command line instead. */
	    for (lineptr = buffer; *lineptr != '%'; lineptr++);
	    sscanf (lineptr + 1, "%s", _STR);
	    if (*(lineptr + 1) != '%') loadfontfile(_STR);
	 }
	 else if (!strcmp(keyword, "loadlibrary")) {
	    /* Deprecated, but retained for backward compatibility */
	    /* Load from script, .xcircuitrc, or command line instead. */
	    int ilib, tlib;

	    for (lineptr = buffer; *lineptr != '%'; lineptr++);
	    sscanf (++lineptr, "%s", _STR);
	    while (isspace(*lineptr)) lineptr++;
	    while (!isspace(*++lineptr));
	    while (isspace(*++lineptr));
	    if (sscanf (lineptr, "%d", &ilib) > 0) {
	       while ((ilib - 2 + LIBRARY) > xobjs.numlibs) {
		  tlib = createlibrary();
		  if (tlib != xobjs.numlibs - 2 + LIBRARY) {
		     ilib = tlib;
		     break;
		  }
	       }
	       mode = ilib - 1 + LIBRARY;
	    }
	    loadlibrary(mode);
	 }
	 else if (!strcmp(keyword, "beginparm")) { /* parameterized object */
	    short tmpnum, i;
	    for (--keyptr; *keyptr == ' '; keyptr--);
	    for (; *keyptr != ' '; keyptr--);
	    sscanf(keyptr, "%hd", &tmpnum);
	    lineptr = buffer;
	    while (isspace(*lineptr)) lineptr++;

	    if (tmpnum < 256) {		/* read parameter defaults in order */
	       stringpart *newpart;

	       localdata->num_params = (u_char) tmpnum;
	       localdata->params = (oparamptr *)malloc(tmpnum * sizeof(oparamptr));
	       for (i = 0; i < tmpnum; i++) {
		  oparamptr ops;
		  *(localdata->params + i) = (oparamptr)malloc(sizeof(oparam));
 		  ops = *(localdata->params + i);
		  if (*lineptr == '(' || *lineptr == '{') {  /* type is XC_STRING */
		     char *linetmp, csave;
		     
		     ops->parameter.string = NULL;

		     /* get simple substring or set of substrings and commands */
		     linetmp = find_match(lineptr);
		     csave = *(++linetmp);
		     *linetmp = '\0';
		     if (*lineptr == '{') lineptr++;
		     readlabel(localdata, lineptr, &(ops->parameter.string));

		     /* Add the ending part to the parameter string */
		     newpart = makesegment(&(ops->parameter.string), NULL);  
		     newpart->type = PARAM_END;

		     ops->type = (u_char)XC_STRING;
		     /* printf("Parameter %d to object %s defaults to string \"%s\"\n",
				i + 1, localdata->name, ops->parameter.string); */
		     *linetmp = csave;
		     lineptr = linetmp;
		     while (isspace(*lineptr)) lineptr++;
		  }
		  else {	/* type is XC_SHORT */
		     ops->type = (u_char)XC_SHORT;
	             sscanf(lineptr, "%hd", &ops->parameter.svalue);
		     /* printf("Parameter %d to object %s defaults to value %hd\n",
				i + 1, localdata->name, ops->parameter.svalue)); */
		  }
	       }
	    }
	 }
#ifdef SCHEMA
	 else if (!strcmp(keyword, "trivial")) {
	    localdata->schemtype = TRIVIAL;
	    if (!areastruct.schemon) doxschema(NULL, NULL, NULL);
	 }
#endif
         else if (!strcmp(keyword, "begingate")) {
	    localdata->num_params = (u_char) 0;
	    localdata->params = (oparamptr *)NULL;
	 }
         else if (!strcmp(keyword, "EndLib")) break;
	 else if (!strcmp(keyword, "restore"));    /* handled at top */
	 else if (!strcmp(keyword, "grestore"));   /* ignore */
         else if (!strcmp(keyword, "endgate"));    /* also ignore */
	 else if (!strcmp(keyword, "xyarray"));	   /* ignore for now */
         else {
	    objinstptr *newinst;
	    char *tmpptr;
	    Boolean found = False;

	    /* first, make sure this is not a general comment line */
	    /* and return if we have a page boundary	   	   */

	    for (tmpptr = buffer; isspace(*tmpptr); tmpptr++);
	    if (*tmpptr == '%') {
	       if (strstr(buffer, "%%Page:") == tmpptr) {
                  strncpy(retstr, buffer, 150);
                  retstr[149] = '\0';
		  free (buffer);
		  return True;
	       }
	       continue;
	    }

	    /* (Assume that this line calls an object instance) */
	    /* Double loop through user libraries 		*/

            for (k = 0; k < ((mode == FONTLIB) ? 1 : xobjs.numlibs); k++) {
	       for (j = 0; j < ((mode == FONTLIB) ? xobjs.fontlib.number :
			xobjs.userlibs[k].number); j++) {
		  libobj = (mode == FONTLIB) ? xobjs.fontlib.library + j :
			xobjs.userlibs[k].library + j;

	          if (!strcmp(keyword, (*libobj)->name)) {
		     char *arrayptr, *endptr;
		     found = True;
		     NEW_OBJINST(newinst, localdata);
      	             localdata->parts++;

	             sscanf(buffer, "%f %hd %hd %hd", &(*newinst)->scale,
		        &(*newinst)->rotation, &(*newinst)->position.x,
		        &(*newinst)->position.y); 
		     (*newinst)->position.x -= offx;
		     (*newinst)->position.y -= offy;
                     if ((*newinst)->rotation > 0)
                        (*newinst)->rotation = (360 - (*newinst)->rotation) /
                           ROT_UNIT + 1;
                     else if ((*newinst)->rotation == 0) (*newinst)->rotation++;
                     else
                        (*newinst)->rotation = ((*newinst)->rotation + 1) /
                            ROT_UNIT - 1;
		     (*newinst)->thisobject = *libobj;
		     (*newinst)->color = curcolor;
		     (*newinst)->params = NULL;

		     /* Does this instance contain parameters? */

		     if ((arrayptr = strchr(buffer, '[')) != NULL) {
		        (*newinst)->params = (oparamptr *)malloc(
				(*libobj)->num_params * sizeof(oparamptr));
		        endptr = find_match(arrayptr);
		        arrayptr++;
		        for (i = 0; i < (*libobj)->num_params; i++) {
			   /* move to next non-space token */
			   while (isspace(*arrayptr) && *arrayptr != '\0')
			      arrayptr++;
			   if (*arrayptr == '\0' || arrayptr == endptr)
			      (*newinst)->params[i] = NULL;
			   else {
			      (*newinst)->params[i] = (oparamptr)malloc(sizeof(oparam));
			      if (*arrayptr == '(' || *arrayptr == '{') { 
			         char *substrend, csave;
				 stringpart *endpart;

			         /* type XC_STRING */

				 (*newinst)->params[i]->parameter.string = NULL;
			         substrend = find_match(arrayptr);
				 csave = *(++substrend);
				 *substrend = '\0';
				 if (*arrayptr == '{') arrayptr++;
				 readlabel(*libobj, arrayptr,
					&((*newinst)->params[i]->parameter.string));

				 *substrend = csave;
			         arrayptr = substrend;

			         /* Append a PARAM_END to the parameter string */

			         endpart = makesegment(
					&((*newinst)->params[i]->parameter.string),
					NULL);
				 endpart->type = PARAM_END;
			      }
			      else {

			         /* type XC_SHORT */

			         sscanf(arrayptr, "%hd",
					&((*newinst)->params[i]->parameter.svalue));
			      /* printf("Object %s called with parameter %d value %hd\n",
					(*newinst)->thisobject->name, i + 1,
					*((short *)((*newinst)->params[i]))); */
			      }
			   }
		        }
		     }
		     break;
	          } /* if !strcmp */
	       } /* for j loop */
	       if (found) break;
	    } /* for k loop */
	    if (!found)		/* will assume that we have a continuation line */
	       temp = continueline(&buffer);
         }
      }
   }
   strncpy(retstr, buffer, 150);
   retstr[149] = '\0';
   free(buffer);
   return True;
}

/*------------------------*/
/* Save a PostScript file */
/*------------------------*/

void setfile(Widget button, Widget fnamewidget, caddr_t clientdata) 
{
   /* see if name has been changed in the buffer */

   sprintf(_STR2, "%.249s", (char *)XwTextCopyBuffer(fnamewidget));
   if (strcmp(xobjs.pagelist[areastruct.page]->filename, _STR2)) {
      Wprintf("Changing name of edit file.");
      xobjs.pagelist[areastruct.page]->filename = (char *) realloc (
	   xobjs.pagelist[areastruct.page]->filename, (strlen(_STR2) + 1) *
	   sizeof(char));
      strcpy(xobjs.pagelist[areastruct.page]->filename, _STR2);
   }
   if (strstr(xobjs.pagelist[areastruct.page]->filename, "Page ") != NULL) {
      Wprintf("Warning: Enter a new name.");
      if (beeper) XBell(dpy, 100);
   }
   else {

      Arg wargs[1];
      Widget db, di;

      savefile(objectdata, CURRENT_PAGE); 

      /* Change "close" button to read "done" */

      di = XtParent(button);
      db = XtNameToWidget(di, "Close");
      XtSetArg(wargs[0], XtNlabel, "  Done  ");
      XtSetValues(db, wargs, 1);
      if (beeper) XBell(dpy, 100);
   }
}

/*--------------------------------------------------------------*/
/* tempfile save						*/
/*--------------------------------------------------------------*/

XtTimerCallbackProc savetemp(Widget w, XtIntervalId *id)
{
   if (xobjs.tempfile == NULL)
   {
      int fd;
      char template[]="/tmp/XCXXXXXX";

      fd = mkstemp(template);
      if (fd == -1) {
	 fprintf(stderr, "Error generating file for savetemp\n");
	 return NULL;
      } 
      close(fd);
      xobjs.tempfile = strdup(template);
   }

   savefile(objectdata, ALL_PAGES);
   xobjs.timeout_id = XtAddTimeOut((u_long)60000 *
	xobjs.save_interval, (XtTimerCallbackProc)savetemp, NULL);
   return NULL;
}

/*--------------------------------------------------------------*/

void savelib(Widget button, caddr_t clientdata, caddr_t nulldata)
{
   buttonsave *savebutton = (buttonsave *)malloc(sizeof(buttonsave));
   int ilib;

   if ((ilib = is_library(objectdata)) < 0) ilib = xobjs.numlibs - 1;

   if (xobjs.userlibs[ilib].number == 0) {
      Wprintf("No user objects to save.");
      return;
   }
   getgeneric(savebutton, button, savelib, (void *)ilib);
   popupprompt(button, "Enter name for library:", "\0", savelibrary,
	savebutton, False);
}

/*---------------------------------------------------------*/

void savelibrary(Widget w, int ilib)
{
   FILE *ps;
   char outname[150], *outptr;
   objectptr *wroteobjs, *libptr;
   short written = 0;
   char *uname = NULL;
   char *hostname = NULL;
   struct passwd *mypwentry = NULL;

   sscanf(_STR2, "%s", outname);
   if ((outptr = strrchr(outname, '/')) == NULL) outptr = outname;
   if (strchr(outptr, '.') == NULL) sprintf(outname, "%s.lps", _STR2);
   tilde_expand(outname);

   ps = fopen(outname, "w");
   if (ps == NULL) {
      Wprintf("Can't open PS file.");
      return;
   }

   fprintf(ps, "%%! PostScript set of library objects for XCircuit\n");
   fprintf(ps, "%%  Version: %2.1f\n", version);
   fprintf(ps, "%%  Library name is: %s\n", _STR2);
   uname = getenv((const char *)"USER");
   if (uname != NULL) mypwentry = getpwnam(uname);

   /* Check for both $HOST and $HOSTNAME environment variables.  Thanks */
   /* to frankie liu <frankliu@Stanford.EDU> for this fix.		*/
   
   if ((hostname = getenv((const char *)"HOSTNAME")) == NULL)
      if ((hostname = getenv((const char *)"HOST")) == NULL) {
	 if (gethostname(_STR, 149) != 0)
	    hostname = uname;
	 else
	    hostname = _STR;
      }

   if (mypwentry != NULL)
      fprintf(ps, "%%  Author: %s <%s@%s>\n", mypwentry->pw_gecos, uname,
		hostname);
   fprintf(ps, "%%\n\n%% XCircuitLib library objects\n");

   /* list of library objects already written */

   wroteobjs = (objectptr *) malloc (sizeof(objectptr));

   /* write all of the object definitions used, bottom up */

   for (libptr = xobjs.userlibs[ilib].library; libptr < xobjs.userlibs[ilib].library + 
        xobjs.userlibs[ilib].number; libptr++)
      printobjects(ps, *libptr, 0.0, 0.0, 0.0, 0, 0, &wroteobjs, &written, DEFAULTCOLOR);

   /* and the postlog */

   fprintf(ps, "\n%% EndLib\n");
   fclose(ps);
   sprintf(_STR, "Library %s saved.", outname);
   Wprintf(_STR);

   free(wroteobjs);
}

/*----------------------------------------------------------------------*/
/* Recursive routine to search the object hierarchy for fonts used	*/
/*----------------------------------------------------------------------*/

void findfonts(objectptr writepage, short *fontsused) {
   genericptr *dfp;
   stringpart *chp;
   int findex;

   for (dfp = writepage->plist; dfp < writepage->plist + writepage->parts; dfp++) {
      if (IS_LABEL(dfp)) {
         for (chp = TOLABEL(dfp)->string; chp != NULL; chp = chp->nextpart) {
	    if (chp->type == FONT_NAME) {
	       findex = chp->data.font;
	       if (fontsused[findex] == 0) {
	          fontsused[findex] = 0x8000 | fonts[findex].flags;
	       }
	    }
	 }
      }
      else if (IS_OBJINST(dfp)) {
	 findfonts(TOOBJINST(dfp)->thisobject, fontsused);
      }
   }
}

/*----------------------------------------------------------------------*/
/* Main file saving routine						*/
/*----------------------------------------------------------------------*/

void savefile(objectptr localdata, short mode) 
{
   FILE *ps, *pro;
   char outname[150], temp[150], prologue[150], *fname, *fptr;
   short written = 0, fontsused[256], i, page, multipage, savepage, stcount;
   objectptr *wroteobjs, writepage, *mpages;
   int *mpageno, findex;
   float psscale, psnorm;
   float xmargin = 72.0, ymargin = 72.0;
   int bboxx, bboxy;
   time_t tdate;

   if (mode == CURRENT_PAGE)
      fname = xobjs.pagelist[areastruct.page]->filename;
   else {
      /* doubly-protected backup: protect against errors during file write */
      sprintf(outname, "%sB", xobjs.tempfile);
      rename(xobjs.tempfile, outname);
      fname = xobjs.tempfile;
   }

   if ((fptr = strrchr(fname, '/')) == NULL) fptr = fname;
   if ((mode == CURRENT_PAGE) && (strchr(fptr, '.') == NULL))
      sprintf(outname, "%s.ps", fname);
   else sprintf(outname, "%s", fname);

   tilde_expand(outname);

   ps = fopen(outname, "w");
   if (ps == NULL) {
      Wprintf("Can't open PS file.");
      return;
   }

   /* calculate the bounding box of the drawing */

   /* find margins */

   psscale = getpsscale(xobjs.pagelist[areastruct.page]->outscale, areastruct.page);
   if (xobjs.pagelist[areastruct.page]->pmode & 1) {
      if (xobjs.pagelist[areastruct.page]->orient == 90) {
	 xmargin = (xobjs.pagelist[areastruct.page]->pagesize.x -
		((float)localdata->height * psscale)) / 2;
	 ymargin = (xobjs.pagelist[areastruct.page]->pagesize.y -
		((float)localdata->width * psscale)) / 2;
      }
      else {
         xmargin = (xobjs.pagelist[areastruct.page]->pagesize.x -
		((float)localdata->width * psscale)) / 2;
         ymargin = (xobjs.pagelist[areastruct.page]->pagesize.y -
		((float)localdata->height * psscale)) / 2;
      }
   }

#ifdef SCHEMA
   /* find all the schematics which are subcircuits of the top-level-page */
   /* circuit and give them the same filename, so that they will be saved */
   /* together as a set.						  */

   if (areastruct.schemon) {
      Wprintf("Gathering all subcircuits. . .");
      findsubschems(areastruct.page, objectdata);
   }
#endif

   /* check for multiple-page output: get the number of pages */

   multipage = 0;
   mpages = (objectptr *) malloc (sizeof(objectptr));
   mpageno = (int *) malloc (sizeof(int));
   savepage = areastruct.page;

   for (page = 0; page < xobjs.pages; page++) {
      writepage = xobjs.pagelist[page]->pageobj;
      if (writepage != NULL)
	 if ((mode == ALL_PAGES) || (xobjs.pagelist[page]->filename
			== xobjs.pagelist[areastruct.page]->filename)) {
	    multipage++;
	    mpages = (objectptr *)realloc(mpages, multipage * sizeof(objectptr));
	    mpageno = (int *)realloc(mpageno, multipage * sizeof(int));
	    mpages[multipage - 1] = writepage;
	    mpageno[multipage - 1] = page;
	 }
   }
   if (multipage == 0) {
      Wprintf("Panic:  could not find this page in page list!");
      free (mpages);
      free (mpageno);
      fclose(ps);
      return;
   }

   /* Print the PostScript DSC Document Header */

   fprintf(ps, "%%!PS-Adobe-3.0");
   if (multipage == 1 && !(xobjs.pagelist[areastruct.page]->pmode & 1))
      fprintf(ps, " EPSF-3.0\n");
   else
      fprintf(ps, "\n");
   fprintf(ps, "%%%%Title: %s\n", xobjs.pagelist[areastruct.page]->filename);
   fprintf(ps, "%%%%Creator: Xcircuit v%2.1f\n", PROG_VERSION);
   tdate = time(NULL);
   fprintf(ps, "%%%%CreationDate: %s", asctime(localtime(&tdate)));
   fprintf(ps, "%%%%Pages: %d\n", multipage);
   if (xobjs.pagelist[areastruct.page]->orient == 90) {
      bboxx = (int)((float)localdata->height * psscale) + (int)xmargin + 4;
      bboxy = (int)((float)localdata->width * psscale) + (int)ymargin + 4;
   }
   else {
      bboxx = (int)((float)localdata->width * psscale) + (int)xmargin + 4;
      bboxy = (int)((float)localdata->height * psscale) + (int)ymargin + 4;
   }
   /* in order to reconstruct the page size, it is necessary that the   */
   /* bounding box extent + margin = page size.  This of course assumes */
   /* that the picture is centered on the page so margins are equal.    */

   if (xobjs.pagelist[areastruct.page]->pmode & 1) 
      fprintf(ps, "%%%%BoundingBox: %d %d %d %d\n",
	   xobjs.pagelist[areastruct.page]->pagesize.x - bboxx,
	   xobjs.pagelist[areastruct.page]->pagesize.y - bboxy, bboxx, bboxy);
   else
      fprintf(ps, "%%%%BoundingBox: %d %d %d %d\n", (int)xmargin - 4,
	   (int)ymargin - 4, bboxx, bboxy);
   for(i = 0; i <= fontcount; i++) fontsused[i] = 0;
   fprintf(ps, "%%%%DocumentNeededResources: ");
   stcount = 27;

   /* find all of the fonts used in this document */
   /* log all fonts which are native PostScript   */

   for (page = 0; page < multipage; page++) {
      writepage = mpages[page];
      findfonts(writepage, fontsused);
   }

   for(i = 0; i <= fontcount; i++) {
      if (fontsused[i] & 0x8000)
	 if ((fonts[i].flags & 0x8018) == 0x0) {
	    stcount += strlen(fonts[i].psname) + 6;
	    if (stcount > OUTPUTWIDTH) {
	       stcount = strlen(fonts[i].psname) + 6;
	       fprintf(ps, "\n%%%%+ ");
	    }
	    fprintf(ps, "font %s ", fonts[i].psname);
	 }
   }

   fprintf(ps, "\n%%%%EndComments\n");

   sprintf(prologue, "%s/%s", PROLOGUE_DIR, PROLOGUE_FILE);
   pro = fopen(prologue, "r");
   if (pro == NULL) {
      sprintf(prologue, "%s", PROLOGUE_FILE);
      pro = fopen(prologue, "r");
      if (pro == NULL) {
         Wprintf("Can't open prolog.");
	 free(mpages);
	 free(mpageno);
	 fclose(ps);
         return;
      }
   }

   /* write the prolog to the output */

   for(;;) {
      if (fgets(temp, 149, pro) == NULL) break;
      fputs(temp, ps);
   }
   fclose(pro);

   /* Special font handling */

   for (findex = 0; findex < fontcount; findex++) {

      /* Derived font slant */

      if ((fontsused[findex] & 0x032) == 0x032)
         fprintf(ps, "/%s /%s .167 fontslant\n\n",
		fonts[findex].psname, fonts[findex].family);

      /* Derived ISO-Latin1 encoding */

      if ((fontsused[findex] & 0xf80) == 0x100) {
	 char *fontorig = NULL;
	 short i;
	 /* find the original standard-encoded font (can be itself) */
	 for (i = 0; i < fontcount; i++) {
	    if (i == findex) continue;
	    if (!strcmp(fonts[i].family, fonts[findex].family) &&
		 ((fonts[i].flags & 0x03) == (fonts[findex].flags & 0x03)))
	       fontorig = fonts[i].psname;
	    if (fontorig == NULL) fontorig = fonts[findex].psname;
	 }
	 fprintf(ps, "/%s findfont dup length dict begin\n", fontorig);
	 fprintf(ps, "{1 index /FID ne {def} {pop pop} ifelse} forall\n");
	 fprintf(ps, "/Encoding ISOLatin1Encoding def currentdict end\n");
	 fprintf(ps, "/%s exch definefont pop\n\n", fonts[findex].psname);
      }

      /* To do:  ISO-Latin2 encoding */

      if ((fontsused[findex] & 0xf80) == 0x180) {
      }

      /* To do:  Special encoding */

      if ((fontsused[findex] & 0xf80) == 0x80) {
      }

      /* To do:  Vectored (drawn) font */

      if (fontsused[findex] & 0x8) {
      }
   }

   fprintf(ps, "%% XCircuit output starts here.\n\n");

   /* list of objects already written---maximum possible = # of builtins */

   wroteobjs = (objectptr *) malloc (sizeof(objectptr));

   for (page = 0; page < multipage; page++) {
      writepage = mpages[page];

      psnorm = xobjs.pagelist[mpageno[page]]->outscale;
      psscale = getpsscale(psnorm, mpageno[page]);
      if (xobjs.pagelist[mpageno[page]]->pmode & 1) {
         if (xobjs.pagelist[mpageno[page]]->orient == 90) {
	    xmargin = (xobjs.pagelist[areastruct.page]->pagesize.x -
		((float)writepage->height * psscale)) / 2;
	    ymargin = (xobjs.pagelist[areastruct.page]->pagesize.y -
		((float)writepage->width * psscale)) / 2;
         }
         else {
            xmargin = (xobjs.pagelist[areastruct.page]->pagesize.x -
		((float)writepage->width * psscale)) / 2;
            ymargin = (xobjs.pagelist[areastruct.page]->pagesize.y -
		((float)writepage->height * psscale)) / 2;
         }
      }

      /* write all of the object definitions used, bottom up */

      printobjects(ps, writepage, psnorm, xmargin, ymargin, page + 1,
	    mpageno[page], &wroteobjs, &written, DEFAULTCOLOR);
      fprintf(ps, "pgsave restore showpage\n\n");
      fflush(ps);
   }
   free (wroteobjs);
   free (mpages);

   /* Done! */

   fprintf(ps, "%%%%Trailer\n");
   fprintf(ps, "XCIRCsave restore\n");
   fprintf(ps, "%%%%EOF\n");
   fclose(ps);

   sprintf(_STR, "File %s saved (%d page%s).", fname, multipage,
		(multipage > 1 ? "s" : ""));
   Wprintf(_STR);

   /* remove the temporary redundant backup */
   if (mode == ALL_PAGES) {
      sprintf(outname, "%sB", xobjs.tempfile);
      unlink(outname);
   }
}

/*----------------------------------------------------------------------*/
/* Given a color index, print the R, G, B values			*/
/*----------------------------------------------------------------------*/

int printRGBvalues(char *tstr, int index, const char *postfix)
{
   int i;
   for (i = 0; i < number_colors; i++) {
      if (colorlist[i].color.pixel == index) {
	 sprintf(tstr, "%4.3f %4.3f %4.3f %s",
		   (float)colorlist[i].color.red   / 65535,
		   (float)colorlist[i].color.green / 65535,
		   (float)colorlist[i].color.blue  / 65535,
		   postfix);
	 return 0;
      }
   }

   /* The program can reach this point for any color which is	*/
   /* not listed in the table.  For instance, the "infolabel" 	*/
   /* green color is not in the list.				*/

   return -1;
}

/*----------------------------------------------------*/
/* Write string to PostScript string, ignoring NO_OPs */
/*----------------------------------------------------*/

int nosprint(char *sptr)
{
   u_char *pptr, *qptr = _STR;
   int retval = 0;

   *qptr++ = '(';

   /* Includes extended character set (non-ASCII) */

   for (pptr = sptr; pptr && *pptr != '\0'; pptr++) {
      if (*pptr > (char)126) {
	 sprintf(qptr, "\\%3o", (int)(*pptr));
	 qptr += 4; 
      }
      else {
         if ((*pptr == '(') || (*pptr == ')') || (*pptr == '\\'))
	    *qptr++ = '\\';
         *qptr++ = *pptr;
      }
   }
   if (qptr == (u_char *)_STR + 1) {
      qptr--;
      retval = -1;
   }
   else {
      *qptr++ = ')';
      *qptr++ = ' ';
   }
   *qptr++ = '\0';

   return retval;
}

/*--------------------------------------------------------------*/
/* Write label segments to the output (in reverse order)	*/
/*--------------------------------------------------------------*/

short writelabel(FILE *ps, stringpart *chrtop, short *stcount)
{
   short i, segs = 0;
   stringpart *chrptr;
   char **ostr = (char **)malloc(sizeof(char *));
   float lastscale = 1.0;
   int lastfont = -1;

   /* Write segments into string array, in forward order */

   for (chrptr = chrtop; chrptr != NULL; chrptr = chrptr->nextpart) {
      ostr = (char **)realloc(ostr, (segs + 1) * sizeof(char *));
      if (chrtop->type == PARAM_END) {	/* NULL parameter is empty string */
	 ostr[segs] = (char *)malloc(3);
	 strcpy(ostr[segs], "()");
      }
      else {
         writesegment(chrptr, &lastscale, &lastfont);
         ostr[segs] = (char *)malloc(1 + strlen(_STR));
         strcpy(ostr[segs], _STR);
      }
      segs++;
   }

   /* Write string array to output in reverse order */
   for (i = segs - 1; i >= 0; i--) {
      dostcount(ps, stcount, strlen(ostr[i]));
      fputs(ostr[i], ps);
      free(ostr[i]);
   }
   free(ostr);	 

   return segs;
}

/*--------------------------------------------------------------*/
/* Write a single label segment to the output			*/
/* (Recursive, so we can write segments in the reverse order)	*/
/*--------------------------------------------------------------*/

void writesegment(stringpart *chrptr, float *lastscale, int *lastfont)
{
   int type = chrptr->type;
   switch(type) {
      case PARAM_START:
	 sprintf(_STR, "v%d ", chrptr->data.paramno + 1);
	 break;
      case PARAM_END:
	 _STR[0] = '\0';
	 chrptr->nextpart = NULL;
	 break;
      case SUBSCRIPT:
	 sprintf(_STR, "{ss} ");
	 break;
      case SUPERSCRIPT:
	 sprintf(_STR, "{Ss} ");
	 break;
      case NORMALSCRIPT:
	 *lastscale = 1.0;
         sprintf(_STR, "{ns} ");
         break;
      case UNDERLINE:
         sprintf(_STR, "{ul} ");
         break;
      case OVERLINE:
         sprintf(_STR, "{ol} ");
         break;
      case NOLINE:
         sprintf(_STR, "{} ");
         break;
      case HALFSPACE:
         sprintf(_STR, "{hS} ");
         break;
      case QTRSPACE:
	 sprintf(_STR, "{qS} ");
	 break;
      case RETURN:
	 *lastscale = 1.0;
	 sprintf(_STR, "{CR} ");
	 break;
      case TABSTOP:
	 sprintf(_STR, "{Ts} ");
	 break;
      case TABFORWARD:
	 sprintf(_STR, "{Tf} ");
	 break;
      case TABBACKWARD:
	 sprintf(_STR, "{Tb} ");
	 break;
      case FONT_NAME:
	 if (*lastscale == 1.0)
	    sprintf(_STR, "{/%s cf} ", fonts[chrptr->data.font].psname);
	 else
	    sprintf(_STR, "{/%s %5.3f cf} ", fonts[chrptr->data.font].psname,
		*lastscale);
	 *lastfont = chrptr->data.font;
	 break;
      case FONT_SCALE:
	 if (*lastfont == -1) {
	    fprintf(stderr, "Warning:  Font may not be the one that was intended.\n");
	    *lastfont = 0;
	 }
	 *lastscale = chrptr->data.scale;
	 sprintf(_STR, "{/%s %5.3f cf} ", fonts[*lastfont].psname, *lastscale);
	 break;
      case FONT_COLOR:
	 strcpy(_STR, "{");
	 if (chrptr->data.color == DEFAULTCOLOR)
	    strcat(_STR, "sce} ");
	 else
	    if (printRGBvalues(_STR + 1,
		    colorlist[chrptr->data.color].color.pixel, "scb} ") < 0)
	       strcat(_STR, "sce} ");
	 break;
      case KERN:
	 sprintf(_STR, "{%d %d Kn} ", chrptr->data.kern[0], chrptr->data.kern[1]);
	 break;
      case TEXT_STRING:
	 nosprint(chrptr->data.string);
	 break;
   }
}

/*--------------------------------------------------------------*/
/* Routine to write all the label segments as stored in _STR	*/
/*--------------------------------------------------------------*/

void writelabelsegs(FILE *ps, short *stcount, stringpart *chrptr)
{
   Boolean ismultipart = ((chrptr->nextpart != NULL) &&
	   (chrptr->nextpart->type != PARAM_END)) ? True : False;

   /* nextpart is not NULL if there are multiple parts to the string */
   if (ismultipart) {
      fprintf(ps, "{");
      (*stcount)++;
   }
   writelabel(ps, chrptr, stcount);

   if (ismultipart) {
      fprintf(ps, "}");
      (*stcount)++;
   }
}

/*--------------------------------------------------------------------------*/
/* recursive routine to print out the library objects used in this drawing, */
/* starting at the bottom of the object hierarchy so that each object is    */
/* defined before it is called.  A list of objects already written is 	    */
/* maintained so that no object is written twice. 			    */
/*--------------------------------------------------------------------------*/

void printobjects(FILE *ps, objectptr localdata, float psnorm, float xmargin,
	float ymargin, short page, short mpage, objectptr **wrotelist,
	short *written, int ccolor)
{
   genericptr *gptr, *savegen;
   objectptr *optr;
   pointlist savept;
   Boolean already;
   short stcount;
   int xaddin, yaddin;
   int curcolor = ccolor;
   float psscale = getpsscale(psnorm, mpage);

   /* first, get a total count of all objects and give warning if large */

   if (localdata->parts > 255) {
      sprintf(_STR, "Warning: \"%s\" may exceed printer's PS limit for definitions",
	    localdata->name);
      Wprintf(_STR);
   }
	   
#ifdef SCHEMA
#define height height2
#define _width width2
#else
#define _width width
#endif

   /* search for all object definitions needed */

   for (gptr = localdata->plist; gptr < localdata->plist + localdata->parts; gptr++)
      if (IS_OBJINST(gptr))
         printobjects(ps, TOOBJINST(gptr)->thisobject, 0.0, 0.0, 0.0, 0, 0,
		wrotelist, written, curcolor);

   if (psscale != 0) {  /* on topmost object of page */
      char *rootptr = strrchr(xobjs.pagelist[mpage]->filename, '/');
      if (rootptr == NULL) rootptr = xobjs.pagelist[mpage]->filename;
      else rootptr++;

      /* If the page label is just the root name of the file, or has been left   */
      /* as "Page n" or "Page_n", just do the normal page numbering.  Otherwise, */
      /* write out the page label explicitly.					 */

      if ((!strcmp(rootptr, localdata->name)) || (strchr(localdata->name, ' ')
		!= NULL) || (strstr(localdata->name, "Page_") != NULL))
         fprintf (ps, "%%%%Page: %d %d\n", page, page);
      else
         fprintf (ps, "%%%%Page: %s %d\n", localdata->name, page);

      if (xobjs.pagelist[mpage]->orient == 90)
         fprintf (ps, "%%%%PageOrientation: Landscape\n");
      else
         fprintf (ps, "%%%%PageOrientation: Portrait\n");
     
      fprintf (ps, "/pgsave save def bop\n");

#ifdef SCHEMA
      if (localdata->symschem != NULL) {
	 fprintf(ps, "%% %s is_symbol\n", localdata->symschem->name);
      }
#endif

      xaddin = (int)(xmargin / psscale) - localdata->lowerleft.x;
      yaddin = (int)(ymargin / psscale) - localdata->lowerleft.y;
      fprintf(ps, "%% %hd %hd offsets\n", xaddin, yaddin);

      if (xobjs.pagelist[mpage]->drawingscale.x != 1
		|| xobjs.pagelist[mpage]->drawingscale.y != 1)
         fprintf(ps, "%% %hd:%hd drawingscale\n", xobjs.pagelist[mpage]->drawingscale.x,
	 	xobjs.pagelist[mpage]->drawingscale.y);

      if (xobjs.pagelist[mpage]->gridspace != 32
		|| xobjs.pagelist[mpage]->snapspace != 16)
         fprintf(ps, "%% %4.2f %4.2f gridspace\n", xobjs.pagelist[mpage]->gridspace,
		xobjs.pagelist[mpage]->snapspace);

      if (xobjs.pagelist[mpage]->background.name != (char *)NULL) {
	 float iscale = (xobjs.pagelist[mpage]->coordstyle == CM) ? CMSCALE : INCHSCALE;
	 fprintf(ps, "%d %d insertion\n",
		(int)((float)xaddin * iscale), (int)((float)yaddin * iscale));
	 savebackground(ps, xobjs.pagelist[mpage]->background.name);
	 fprintf(ps, "end_insert\n");
      }

      if (xobjs.pagelist[mpage]->orient == 90)
         fprintf(ps, "90 rotate %d %d translate\n", (int)(ymargin - xmargin),
	     -((int)((float)localdata->height * psscale) + 
	     (int)(xmargin + ymargin)));

      fprintf(ps, "%5.4f ", psnorm);
      switch(xobjs.pagelist[mpage]->coordstyle) {
	 case FRAC_INCH: case DEC_INCH:
	    fprintf(ps, "inchscale\n");
	    break;
	 case CM:
	    fprintf(ps, "cmscale\n");
	    break;
      };
      fprintf(ps, "%5.4f setlinewidth\n\n", 1.3 *
      	     xobjs.pagelist[mpage]->wirewidth);
   }

   /* search among the list of objects already written to the output */

   already = False;
   for (optr = *wrotelist; optr < *wrotelist + *written; optr++)
      if (*optr == localdata) {
	  already = True;
	  break;
      }

   if (!already) {

      /* update the list of objects already written to the output */
      if (psscale == 0) {
	 *wrotelist = (objectptr *)realloc(*wrotelist, (*written + 1) * 
		sizeof(objectptr));
         *(*wrotelist + *written) = localdata;
         (*written)++;
      }

      if (psscale == 0) {
	 fprintf(ps, "/%s {\n", localdata->name);
#ifdef SCHEMA
	 sprintf(_STR, "%% %hd %hd %hd %hd bbox\n", localdata->lleft2.x,
	      localdata->lleft2.y, localdata->_width, localdata->height);
#else
	 sprintf(_STR, "%% %hd %hd %hd %hd bbox\n", localdata->lowerleft.x,
	      localdata->lowerleft.y, localdata->_width, localdata->height);
#endif
	 fputs(_STR, ps);
	 if (localdata->hidden == True) fprintf(ps, "%% hidden\n");

#ifdef SCHEMA
	 /* For symbols with schematics, and "trivial" schematics */
         if (areastruct.schemon) {
            if (localdata->symschem != NULL)
	       fprintf(ps, "%% %s is_schematic\n", localdata->symschem->name);
            else if (localdata->schemtype == TRIVIAL)
	       fprintf(ps, "%% trivial\n");
         }
#endif

	 /* Check for parameters and default values */
         if (localdata->params != NULL) {
	    short i;
	    stcount = 0;
	    for (i = 0; i < localdata->num_params; i++) {
	       if (localdata->params[i]->type == XC_STRING) {
		  writelabelsegs(ps, &stcount, localdata->params[i]->parameter.string);
		  fprintf(ps, " ");
	       }
	       else {	/* XC_SHORT */
		  fprintf(ps, "%hd ", localdata->params[i]->parameter.svalue);
	       }
	       stcount++;
	    }
	    sprintf(_STR, "%d beginparm\n", localdata->num_params);
	    dostcount (ps, &stcount, strlen(_STR));
	    fputs(_STR, ps);
	 }
	 else 
	    fprintf(ps, "begingate\n");
      }

      if (psscale == 0) {
	 xaddin = yaddin = 0;
      }


#ifdef SCHEMA
#undef height
#endif
#undef _width

      /* write all the elements in order */

      for (savegen = localdata->plist; savegen < localdata->plist +
	   localdata->parts; savegen++) {

	 /* change current color if different */

	 if ((*savegen)->color != curcolor) {
	    if ((curcolor = (*savegen)->color) == DEFAULTCOLOR)
	       fprintf(ps, "sce\n");
	    else {
	       if (printRGBvalues(_STR, (*savegen)->color, "scb\n") < 0) {
		  fprintf(ps, "sce\n");
		  curcolor = DEFAULTCOLOR;
	       }
	       else
	          fputs(_STR, ps); 
	    }
	 }

	 stcount = 0;
	 switch((*savegen)->type) {

	    case(POLYGON):
               sprintf(_STR, "%hd %5.2f ", TOPOLY(savegen)->style,
		     TOPOLY(savegen)->width);
	       stcount = strlen(_STR);
	       fputs(_STR, ps);
               for (savept = TOPOLY(savegen)->points; savept < TOPOLY(savegen)->
		      points + TOPOLY(savegen)->number; savept++) {
		  xypmat(savept);
	       }
	       sprintf(_STR, "%hd polygon\n", TOPOLY(savegen)->number);
	       dostcount (ps, &stcount, strlen(_STR));
	       fputs(_STR, ps);
	       break;

	    case(PATH): {
	       genericptr *pgen;
	       pgen = TOPATH(savegen)->plist;
	       switch((*pgen)->type) {
		  case ARC:
		     xyfmat(TOARC(pgen)->points[0]);
		     break;
		  case POLYGON:
		     xymat(TOPOLY(pgen)->points[0]);
		     break;
		  case SPLINE:
		     xyfmat(TOSPLINE(pgen)->ctrl[0]);
		     break;
	       }
	       dostcount(ps, &stcount, 9);
	       fprintf(ps, "beginpath\n");
	       for (pgen = TOPATH(savegen)->plist; pgen < TOPATH(savegen)->plist
			+ TOPATH(savegen)->parts; pgen++) {
	          switch((*pgen)->type) {
		     case ARC:
			if (abs(TOARC(pgen)->radius) == TOARC(pgen)->yaxis) {
			   xymat(TOARC(pgen)->position);
			   fprintf(ps, "%hd ", abs(TOARC(pgen)->radius));
			   if (TOARC(pgen)->radius < 0)
			      fprintf(ps, "%3.2f %3.2f arcn\n",
			         TOARC(pgen)->angle2, TOARC(pgen)->angle1);
			   else
			      fprintf(ps, "%3.2f %3.2f arc\n",
			         TOARC(pgen)->angle1, TOARC(pgen)->angle2);
			}
		        else {
			   xymat(TOARC(pgen)->position);
			   fprintf(ps, "%hd %hd ",
			      abs(TOARC(pgen)->radius), TOARC(pgen)->yaxis);
			   if (TOARC(pgen)->radius < 0)
			      fprintf(ps, "%3.2f %3.2f nellip\n",
			         TOARC(pgen)->angle2, TOARC(pgen)->angle1);
			   else
			      fprintf(ps, "%3.2f %3.2f pellip\n",
			         TOARC(pgen)->angle1, TOARC(pgen)->angle2);
			}
			break;
		     case POLYGON:
               		for (savept = TOPOLY(pgen)->points + TOPOLY(pgen)->number
				- 1; savept > TOPOLY(pgen)->points; savept--) {
			   xypmat(savept);
	       		}
	       		sprintf(_STR, "%hd polyc\n", TOPOLY(pgen)->number - 1);
	       		dostcount (ps, &stcount, strlen(_STR));
	       		fputs(_STR, ps);
	       		break;
		     case SPLINE:
			xymat(TOSPLINE(pgen)->ctrl[1]);
			xymat(TOSPLINE(pgen)->ctrl[2]);
			xymat(TOSPLINE(pgen)->ctrl[3]);
			fprintf(ps, "curveto\n");
			break;
		  }
	       }
	       fprintf(ps, "%hd %5.2f endpath\n", TOPATH(savegen)->style,
			TOPATH(savegen)->width);
	       } break;
	    case(SPLINE):
	       fprintf(ps, "%hd %5.2f ",
	          TOSPLINE(savegen)->style, TOSPLINE(savegen)->width);
	       xymat(TOSPLINE(savegen)->ctrl[1]);
	       xymat(TOSPLINE(savegen)->ctrl[2]);
	       xymat(TOSPLINE(savegen)->ctrl[3]);
	       xymat(TOSPLINE(savegen)->ctrl[0]);
	       fprintf(ps, "spline\n");
	       break;

	    case(ARC):
	       if (abs(TOARC(savegen)->radius) == TOARC(savegen)->yaxis) {
                  fprintf(ps, "%hd %5.2f ",
			TOARC(savegen)->style, TOARC(savegen)->width);
		  xymat(TOARC(savegen)->position);
                  fprintf(ps, "%hd %3.2f %3.2f xcarc\n", abs(TOARC(savegen)->radius),
			TOARC(savegen)->angle1, TOARC(savegen)->angle2);
	       }
	       else {
                  fprintf(ps, "%hd %5.2f ",
			TOARC(savegen)->style, TOARC(savegen)->width);
		  xymat(TOARC(savegen)->position);
                  fprintf(ps, "%hd %hd %3.2f %3.2f ellipse\n",
			abs(TOARC(savegen)->radius), TOARC(savegen)->yaxis,
			TOARC(savegen)->angle1, TOARC(savegen)->angle2);
	       }
	       break;

	    case(OBJECT):
	       {objinstptr sobj = TOOBJINST(savegen);
	       short rotation = sobj->rotation;
	       if (rotation < 0) rotation = (rotation + 1) * ROT_UNIT - 1;
	       else rotation = 360 - ((rotation - 1) * ROT_UNIT);
	       if (rotation == 360) rotation = 0;
	       sprintf(_STR, "%3.2f %hd ", sobj->scale, rotation);
	       dostcount(ps, &stcount, strlen(_STR));
	       fputs(_STR, ps);
	       xymat(sobj->position);

	       /* check for parameters */
	       if (sobj->params != NULL) {
		  short i, iend;

		  for (iend = sobj->thisobject->num_params; iend > 0; iend--)
		     if (sobj->params[iend - 1] != NULL) break;

		  if (iend) {
		     fprintf(ps, "[");
		     stcount++;
		  }
		  for (i = 0; i < iend; i++) {
		     if (sobj->thisobject->params[i]->type == XC_STRING) {
			stringpart *tmpptr;

			if (sobj->params[i] == NULL)     /* use the default */
			   tmpptr = sobj->thisobject->params[i]->parameter.string;   
			else
	                   tmpptr = sobj->params[i]->parameter.string;
		        writelabelsegs(ps, &stcount, tmpptr);
		     }
		     else {	/* XC_SHORT */
			if (sobj->params[i] == NULL) {   /* use the default */
			   sprintf(_STR, "%hd ",
				localdata->params[i]->parameter.svalue);
			}
			else {
			   sprintf(_STR, "%hd ", sobj->params[i]->parameter.svalue);
			}
			dostcount(ps, &stcount, strlen(_STR));
	       		fputs(_STR, ps);
		     }
		  }
		  if (iend) {
		     fprintf(ps, "] ");
		     stcount += 2;
		  }
	       }
	       fprintf(ps, "%s\n", TOOBJINST(savegen)->thisobject->name);
	       }break;
            
	    case(LABEL):
	       {short segs, rotation;
	       Boolean has_param = False;

#ifdef SCHEMA
	       /* strip schematic pins and labels if schematic capture is off */
	       if (TOLABEL(savegen)->pin && !areastruct.schemon) break;
#endif

	       /* Check for parameter --- must use "mark" to count # segments */
	       has_param = hasparameter(TOLABEL(savegen));

	       if (has_param) {
		  fprintf(ps, "mark ");
	          stcount += 5;
	       }

	       segs = writelabel(ps, TOLABEL(savegen)->string, &stcount);

	       rotation = 360 - (TOLABEL(savegen)->rotation - 1) * ROT_UNIT;
	       if (rotation == 360) rotation = 0; 
	       if (segs > 0) {
		  if (has_param)
                     sprintf(_STR, "ctmk ");
		  else
                     sprintf(_STR, "%hd ", segs);
		  dostcount(ps, &stcount, strlen(_STR));
	          fputs(_STR, ps);
                  sprintf(_STR, "%hd %hd %3.2f ", TOLABEL(savegen)->justify,
				rotation, TOLABEL(savegen)->scale);
		  dostcount(ps, &stcount, strlen(_STR));
	          fputs(_STR, ps);
		  xymat(TOLABEL(savegen)->position);

#ifdef SCHEMA
		  if (areastruct.schemon) {
		     switch(TOLABEL(savegen)->pin) {
		        case LOCAL:
			   strcpy(_STR, "pinlabel\n"); break;
			case GLOBAL:
			   strcpy(_STR, "pinglobal\n"); break;
			case INFO:
			   strcpy(_STR, "infolabel\n"); break;
			default:
		     	   strcpy(_STR, "label\n");
		     }
		  }
		  else strcpy(_STR, "label\n");
#else
		  strcpy(_STR, "label\n");
#endif
		  dostcount(ps, &stcount, strlen(_STR));
	          fputs(_STR, ps);
	       }
	       }break;
         }

      }

      /* if this is not the top level, end the object definition */

      if (psscale == 0) fprintf(ps, "endgate\n} def\n\n");

   }
}

/*----------------------------------------------------------------------*/
